# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


from pathlib import Path
from typing import List, NamedTuple
import gluonts
import logging
import numpy as np
import os
import pandas as pd
import pytest
import random
import sys
import tempfile
import warnings

from gluonts.model import Predictor
from gluonts.dataset.common import ListDataset

try:
    import mxnet as mx
except (ImportError, OSError):
    mx = None

try:
    import statsmodels
except ImportError:
    statsmodels = None


class HierarchicalMetaData(NamedTuple):
    S: np.ndarray
    freq: str
    nodes: List


class HierarchicalTrainDatasets(NamedTuple):
    train: ListDataset
    test: ListDataset
    metadata: HierarchicalMetaData


@pytest.fixture
def sine7():
    def _sine7(
        seq_length: int = 100,
        prediction_length: int = 10,
        nonnegative: bool = False,
        bias: float = 0,
    ):
        x = np.arange(0, seq_length)

        # Bottom layer (4 series)
        amps = [0.8, 0.9, 1, 1.1]
        freqs = [1 / 20, 1 / 30, 1 / 50, 1 / 100]

        b = np.zeros((4, seq_length))
        for i, f in enumerate(freqs):
            omega = 0
            if i == 3:
                np.random.seed(0)
                omega = np.random.uniform(0, np.pi)  # random phase shift
            b[i, :] = amps[i] * np.sin(2 * np.pi * x * f + omega) + bias

        if nonnegative:
            b = abs(b)

        # Aggregation matrix S
        S = np.array(
            [
                [1, 1, 1, 1],
                [1, 1, 0, 0],
                [0, 0, 1, 1],
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, 1],
            ],
        )

        Y = S @ b

        # Indices and timestamps
        index = pd.period_range(
            start=pd.Period("2020-01-01", freq="D"),
            periods=Y.shape[1],
            freq="D",
        )

        metadata = HierarchicalMetaData(
            S=S, freq=index.freqstr, nodes=[2, [2] * 2]
        )

        train_dataset = ListDataset(
            [
                {
                    "start": index[0],
                    "item_id": "all_items",
                    "target": Y[:, :-prediction_length],
                }
            ],
            freq=index.freqstr,
            one_dim_target=False,
        )

        test_dataset = ListDataset(
            [{"start": index[0], "item_id": "all_items", "target": Y}],
            freq=index.freqstr,
            one_dim_target=False,
        )

        assert Y.shape[0] == S.shape[0]
        return HierarchicalTrainDatasets(
            train=train_dataset, test=test_dataset, metadata=metadata
        )

    return _sine7


@pytest.fixture(scope="function", autouse=True)
def function_scope_seed(request):
    """A function scope fixture that manages rng seeds.
    This fixture automatically initializes the python, numpy and mxnet random
    number generators randomly on every test run.
    def test_ok_with_random_data():
        ...
    To fix the seed used for a test case mark the test function with the
    desired seed:
    @pytest.mark.seed(1)
    def test_not_ok_with_random_data():
        '''This testcase actually works.'''
        assert 17 == random.randint(0, 100)
    When a test fails, the fixture outputs the seed used. The user can then set
    the environment variable MXNET_TEST_SEED to the value reported, then rerun
    the test with:
        pytest --verbose -s <test_module_name.py> -k <failing_test>
    To run a test repeatedly, install pytest-repeat and add the --count argument:
        pip install pytest-repeat
        pytest --verbose -s <test_module_name.py> -k <failing_test> --count 1000
    """

    seed = request.node.get_closest_marker("seed")
    env_seed_str = os.getenv("MXNET_TEST_SEED")

    if seed is not None:
        seed = seed.args[0]
        assert isinstance(seed, int)
    elif env_seed_str is not None:
        seed = int(env_seed_str)
    else:
        seed = int.from_bytes(os.urandom(4), "big")

    post_test_state = np.random.get_state()
    np.random.seed(seed)
    random.seed(seed)

    if mx is not None:
        mx.random.seed(seed)

    seed_message = (
        "np/mx/python random seeds are set to "
        f"{seed}, use MXNET_TEST_SEED={seed} to reproduce."
    )

    # Always log seed on DEBUG log level. This makes sure we can find out the
    # value of the seed even if the test case causes a segfault and subsequent
    # teardown code is not run.
    logging.debug(seed_message)

    yield  # run the test

    np.random.set_state(post_test_state)


@pytest.fixture(autouse=True)
def doctest(doctest_namespace):
    doctest_namespace["np"] = np
    doctest_namespace["gluonts"] = gluonts

    if mx is not None:
        doctest_namespace["mx"] = mx
        doctest_namespace["gluon"] = mx.gluon

    import doctest

    doctest.ELLIPSIS_MARKER = "-etc-"


collect_ignore = []


def pytest_configure(config):
    test_folder = Path(__file__).parent.resolve()

    targets = config.getoption("file_or_dir")

    if not targets:
        targets.append(".")

    requirements = set()

    for target in targets:
        target = Path(target).resolve()

        if target.is_file():
            target = target.parent
        else:
            requirements.update(target.glob("**/require-packages.txt"))

        try:
            # .is_relative_to was only added in Py 3.9
            target.relative_to(test_folder)

            while True:
                require = target / "require-packages.txt"
                if require.exists():
                    requirements.add(require)

                if target == test_folder:
                    break
                target = target.parent
        except ValueError:
            pass

    old_path = sys.path
    sys.path = [
        path for path in sys.path if Path(path).resolve() != test_folder
    ]

    excludes = []

    for path in requirements:
        with path.open() as requirement:
            for requirement in map(str.strip, requirement):
                try:
                    __import__(requirement)
                except ImportError:
                    collect_ignore.append(
                        str(path.parent.relative_to(test_folder))
                    )
                    break

    if collect_ignore:
        warnings.warn(
            f"Skipping tests because some packages are not installed: {excludes}"
        )

    sys.path = old_path


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


def pytest_runtest_setup(item):
    skip_datasets = [
        mark.args[0] for mark in item.iter_markers(name="skip_dataset")
    ]

    if skip_datasets:
        ds_name = item._request.getfixturevalue("dsinfo")["name"]
        if ds_name in skip_datasets:
            pytest.skip(f"Skip test on dataset {ds_name}")


@pytest.fixture(scope="session", params=["synthetic", "constant"])
def dsinfo(request):
    from gluonts import time_feature
    from gluonts.dataset.artificial import constant_dataset, default_synthetic

    if request.param == "constant":
        ds_info, train_ds, test_ds = constant_dataset()

        return AttrDict(
            name="constant",
            cardinality=int(ds_info.metadata.feat_static_cat[0].cardinality),
            freq=ds_info.metadata.freq,
            num_parallel_samples=2,
            prediction_length=ds_info.prediction_length,
            # FIXME: Should time features should not be needed for GP
            time_features=[time_feature.day_of_week, time_feature.hour_of_day],
            train_ds=train_ds,
            test_ds=test_ds,
        )
    elif request.param == "synthetic":
        ds_info, train_ds, test_ds = default_synthetic()

        return AttrDict(
            name="synthetic",
            batch_size=32,
            cardinality=int(ds_info.metadata.feat_static_cat[0].cardinality),
            context_length=2,
            freq=ds_info.metadata.freq,
            prediction_length=ds_info.prediction_length,
            num_parallel_samples=2,
            train_ds=train_ds,
            test_ds=test_ds,
            time_features=None,
        )


def from_hyperparameters(Forecaster, hyperparameters, dsinfo):
    return Forecaster.from_hyperparameters(
        freq=dsinfo.freq,
        **{
            "prediction_length": dsinfo.prediction_length,
            "num_parallel_samples": dsinfo.num_parallel_samples,
        },
        **hyperparameters,
    )


@pytest.fixture()
def accuracy_test(dsinfo):
    from gluonts.evaluation import backtest_metrics, Evaluator

    def test_accuracy(Forecaster, hyperparameters, accuracy):
        forecaster = from_hyperparameters(Forecaster, hyperparameters, dsinfo)

        if isinstance(forecaster, Predictor):
            predictor = forecaster
        else:
            predictor = forecaster.train(training_data=dsinfo.train_ds)

        agg_metrics, item_metrics = backtest_metrics(
            test_dataset=dsinfo.test_ds,
            predictor=predictor,
            evaluator=Evaluator(
                calculate_owa=statsmodels is not None, num_workers=0
            ),
        )

        if dsinfo.name == "synthetic":
            accuracy = 10.0

        assert agg_metrics["ND"] <= accuracy

    return test_accuracy


@pytest.fixture()
def serialize_test(dsinfo):
    from gluonts.model.predictor import Predictor

    def test_serialize(Forecaster, hyperparameters):
        forecaster = from_hyperparameters(Forecaster, hyperparameters, dsinfo)

        if isinstance(forecaster, Predictor):
            predictor_act = forecaster
        else:
            predictor_act = forecaster.train(dsinfo.train_ds)

        with tempfile.TemporaryDirectory() as temp_dir:
            predictor_act.serialize(Path(temp_dir))
            predictor_exp = Predictor.deserialize(Path(temp_dir))
            # TODO: DeepFactorEstimator does not pass this assert
            assert predictor_act == predictor_exp

    return test_serialize



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass
from typing import List, Tuple

import numpy as np


@dataclass
class ArrowDecoder:
    reshape_columns: List[Tuple[str, str]]

    @classmethod
    def from_schema(cls, schema):
        return cls(
            [
                (column.name[: -len("._np_shape")], column.name)
                for column in schema
                if column.name.endswith("._np_shape")
            ]
        )

    def decode(self, batch, row_number: int):
        return next(self.decode_batch(batch.slice(row_number, row_number + 1)))

    def decode_batch(self, batch):
        for row in batch.to_pandas().to_dict("records"):
            for column_name, shape_column in self.reshape_columns:
                row[column_name] = row[column_name].reshape(
                    row.pop(shape_column)
                )

            for name, value in row.items():
                if type(value) == np.ndarray and value.dtype == object:
                    row[name] = np.stack(value)

            yield row



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.


"""
This modules contains pydantic imports, which are used throughout the codebase.
"""

from pydantic import __version__

PYDANTIC_V2 = __version__.startswith("2")

if PYDANTIC_V2:
    import pydantic.v1 as pydantic
    from pydantic.v1 import (
        BaseConfig,
        BaseModel,
        create_model,
        root_validator,
        PositiveInt,
        PrivateAttr,
        Field,
        parse_obj_as,
        PositiveFloat,
        BaseSettings,
    )
    from pydantic.v1.error_wrappers import ValidationError, display_errors
    from pydantic.v1.utils import deep_update
    from pydantic.v1.dataclasses import dataclass
else:
    import pydantic  # type: ignore[no-redef]
    from pydantic import (  # type: ignore[no-redef, assignment]
        BaseConfig,
        BaseModel,
        create_model,
        root_validator,
        PositiveInt,
        PrivateAttr,
        Field,
        parse_obj_as,
        PositiveFloat,
        BaseSettings,
    )
    from pydantic.error_wrappers import ValidationError, display_errors  # type: ignore[no-redef]
    from pydantic.utils import deep_update  # type: ignore[no-redef]
    from pydantic.dataclasses import dataclass  # type: ignore[no-redef]


__all__ = [
    "BaseConfig",
    "BaseModel",
    "BaseSettings",
    "Field",
    "PositiveFloat",
    "PositiveInt",
    "PrivateAttr",
    "ValidationError",
    "__version__",
    "create_model",
    "dataclass",
    "deep_update",
    "display_errors",
    "parse_obj_as",
    "pydantic",
    "root_validator",
]



import re
import sys
import time
from pathlib import Path

import black
import click
import nbformat
import notedown
from nbclient import NotebookClient

from jinja2 import Environment

env = Environment()


def run_notebook(text, kernel_name, timeout, execute) -> str:
    # We add two blank lines at the end to ensure
    # that the final cell also runs.
    text += "\n" * 2
    notebook = notedown.MarkdownReader().reads(text)

    if execute:
        kwargs = {}
        if kernel_name is not None:
            kwargs["kernel_name"] = kernel_name

        client = NotebookClient(
            notebook,
            timeout=timeout,
            resources={"metadata": {"path": "."}},
            **kwargs,
        )
        client.execute()

    # need to add language info to for syntax highlight
    notebook["metadata"].update(language_info={"name": "python"})

    return nbformat.writes(notebook)


def black_cells(text):
    CODE_RE = r"```py(?:thon)?\s*\n(.*?)```"

    text = re.sub(r"^%", r"# %-% #", text, flags=re.M)

    def apply_black(match):
        code = match.group(1)

        formatted = black.format_str(code, mode=black.Mode())

        return "\n".join(["```", formatted.rstrip(), "```"])

    formatted = re.sub(CODE_RE, apply_black, text, flags=re.S)
    return re.sub(r"^# %-% #", r"%", formatted, flags=re.M)


def convert(path, mode, kernel_name=None, timeout=40 * 60):
    print(f"=== {path.name} ", end="")
    sys.stdout.flush()

    with path.open() as in_file:
        template = env.from_string(in_file.read())

    markdown = template.render(mode=mode)
    markdown = black_cells(markdown)

    start = time.time()
    output = run_notebook(
        markdown,
        kernel_name,
        timeout,
        execute=mode != "skip",
    )
    print(f"finished evaluation in {time.time() - start} sec")

    # XXX.md.input -> XXX.ipynb
    # `with_suffix` only operates on last suffix, so we need some more involved
    # logic.
    stem = path.name.split(".", 1)[0]
    with path.with_name(stem).with_suffix(".ipynb").open("w") as outfile:
        outfile.write(output)


@click.command()
@click.argument("paths", type=click.Path(), nargs=-1)
@click.option("--kernel", "-k", help="Name of iPython kernel to use.")
@click.option("--mode", "-m", default="release")
def cli(paths, kernel, mode):
    for file in map(Path, paths):
        convert(file, kernel_name=kernel, mode=mode)


if __name__ == "__main__":
    cli()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
This module contains functions that work on ``Optional`` values. It supports
wrapping of values into a dedicated type (``Maybe``), but also works on normal
Python values, which are of type ``Optional[T]``.

Each function is implemented twice, as a simple function and as a method on
``maybe.Maybe``::

    maybe.Some(1).map(fn) -> Maybe[T]

    maybe.map(1, fn) -> Optional[T]


Methods on ``Maybe`` return ``Maybe`` types, while functions return
``Optional`` values.

The names are taken from Rust, see:
https://doc.rust-lang.org/stable/std/option/enum.Option.html

.. note::
    The argument order for ``map_or`` and ``map_or_else`` is reversed
    compared to their Rust counterparts.

    ``do`` is not implemented in Rust but mimics ``toolz.do`` instead.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import (
    Callable,
    Generic,
    Optional,
    TypeVar,
    Tuple,
    List,
    Union,
    cast,
    Any,
)
from typing_extensions import final, ParamSpec, Concatenate

T = TypeVar("T")
U = TypeVar("U")
R = TypeVar("R")

P = ParamSpec("P")

OptionalOrMaybe = Union[Optional[T], "Maybe[T]"]


def box(val: OptionalOrMaybe[T]) -> Maybe[T]:
    """
    Turn ``Optional[T]`` into ``Maybe[T]``.
    """

    if isinstance(val, Maybe):
        return val

    if val is None:
        return Nothing

    return Some(val)


def unbox(val: OptionalOrMaybe[T]) -> Optional[T]:
    """
    Turn ``Optional[T]`` into ``Maybe[T]``.
    """

    if isinstance(val, Maybe):
        return val.unbox()

    return val


def flatten(val: Optional[Optional[T]]) -> Optional[T]:
    """
    Flatten nested optional value.

    Note: This just returns the value, but changes the type from
    ``Optional[Optional[T]]`` to ``Optional[T].``
    """

    return val  # type: ignore


def expect(val: OptionalOrMaybe[T], msg: str) -> T:
    """
    Ensure that ``val`` is not ``None``, raises a ``ValueError`` using ``msg``
    otherwise.

    >>> expect(1, "My message")
    1
    >>> expect(None, "My message")
    Traceback (most recent call last):
        ...
    ValueError: My message
    """
    return box(val).expect(msg)


def do(val: OptionalOrMaybe[T], fn: Callable[[T], U]) -> Optional[T]:
    """
    Apply ``fn`` to ``val`` then return ``val``, if ``val`` is not ``None``.

    >>> do("a", print)
    a
    'a'
    >>> do(None, print)
    """
    return box(val).do(fn).unbox()


def map(
    val: OptionalOrMaybe[T],
    fn: Callable[Concatenate[T, P], U],
    *args: P.args,
    **kwargs: P.kwargs,
) -> Optional[U]:
    """
    Apply ``fn`` to ``val`` if ``val`` is not ``None``.

    >>> map(1, lambda x: x + 1)
    2
    >>> map(None, lambda x: x + 1)

    Allows to pass additional arguments that are passed to ``fn``:

    >>> map(10, divmod, 3)
    (3, 1)
    """
    return box(val).map(fn, *args, **kwargs).unbox()


def map_or(val: OptionalOrMaybe[T], fn: Callable[[T], U], default: U) -> U:
    """
    Apply ``fn`` to ``val`` if ``val`` is not ``None`` and return the result.
    In case of ``None`` the provided ``default`` is returned instead.

    This is similar to calling ``map`` and ``unwrap_or`` in succession.

    >>> map_or(["x"], len, 0)
    1
    >>> map_or(None, len, 0)
    0
    """
    return box(val).map_or(fn, default)


def map_or_else(
    val: OptionalOrMaybe[T],
    fn: Callable[[T], U],
    factory: Callable[[], U],
) -> U:
    """
    Similar to ``map_or``, except that the returned value is lazily evaluated.

    This is similar to calling ``map`` and ``unwrap_or_else`` in succession.

    >>> map_or_else(1, lambda n: [n], list)
    [1]
    >>> map_or_else(None, lambda n: [n], list)
    []
    """
    return box(val).map_or_else(fn, factory)


def unwrap(val: OptionalOrMaybe[T]) -> T:
    """
    Assert that the value is not ``None``.

    >>> unwrap(1)
    1
    >>> unwrap(None)
    Traceback (most recent call last):
        ...
    ValueError: Trying to unwrap `None` value.
    """
    return box(val).expect("Trying to unwrap `None` value.")


def unwrap_or(val: OptionalOrMaybe[T], default: T) -> T:
    """
    Get ``val`` if it is not ``None``, or ``default`` otherwise.

    >>> unwrap_or(1, 2)
    1
    >>> unwrap_or(None, 2)
    2
    """
    return box(val).unwrap_or(default)


def unwrap_or_else(val: OptionalOrMaybe[T], factory: Callable[[], T]) -> T:
    """
    Get ``val`` if it is not ``None``, or invoke ``factory`` to get a fallback.

    >>> unwrap_or_else([1, 2, 3], list)
    [1, 2, 3]
    >>> unwrap_or_else(None, list)
    []
    """

    return box(val).unwrap_or_else(factory)


def and_(val: OptionalOrMaybe[T], other: OptionalOrMaybe[U]) -> Optional[U]:
    """
    Like ``a and b`` in Python, except only considering ``None``.

    This is implement identical to ``unwrap_or``, but different with respect to
    types.

    >>> and_(1, 2)
    2
    >>> and_(1, None)
    >>> and_(None, 2)
    """
    return box(val).and_(other).unbox()


def and_then(
    val: OptionalOrMaybe[T],
    fn: Callable[Concatenate[T, P], OptionalOrMaybe[U]],
    *args: P.args,
    **kwargs: P.kwargs,
) -> Optional[U]:
    """
    Apply ``fn`` to ``val`` if it is not ``None`` and return the result.

    In contrast to ``map``, ``fn`` always returns an ``Optional``, which is
    consequently flattened.

    >>> and_then([42], lambda xs: xs[0] if xs else None)
    42
    >>> and_then(None, lambda xs: xs[0] if xs else None)
    """
    return box(val).and_then(fn, *args, **kwargs).unbox()


def or_(val: OptionalOrMaybe[T], default: Optional[T]) -> Optional[T]:
    """
    Like ``a or b`` in Python, except only considering ``None``.

    >>> or_(1, 2)
    1
    >>> or_(1, None)
    1
    >>> or_(None, 2)
    2
    """
    return box(val).or_(default).unbox()


def or_else(
    val: OptionalOrMaybe[T], factory: Callable[[], Optional[T]]
) -> Optional[T]:
    """
    Like ``unwrap_or_else``, except that it returns an optional value.

    >>> or_else([42], list)
    [42]
    >>> or_else(None, list)
    []
    """
    return box(val).or_else(factory).unbox()


def contains(val: OptionalOrMaybe[T], other: U) -> bool:
    """
    Check if ``val`` equals ``other``, always return ``False`` if ``val`` is
    ``None``.

    >>> contains(1, 1)
    True
    >>> contains(1, 2)
    False
    >>> contains(None, 3)
    False
    """
    return box(val).contains(other)


def filter(val: OptionalOrMaybe[T], pred: Callable[[T], bool]) -> Optional[T]:
    """
    Return ``None`` if ``val`` is ``None`` or if ``pred(val)`` does not return
    ``True``, otherwise return ``val``.

    >>> is_even = lambda n: n % 2 == 0
    >>> filter(1, is_even)
    >>> filter(2, is_even)
    2
    >>> filter(None, is_even)
    """
    return box(val).filter(pred).unbox()


def xor(val: OptionalOrMaybe[T], other: OptionalOrMaybe[T]) -> Optional[T]:
    """
    Return either ``val`` or ``other`` if the other is ``None``. Also return
    ``None`` if both are not ``None``.

    >>> xor(1, None)
    1
    >>> xor(None, 2)
    2
    >>> xor(1, 2)
    >>> xor(None, None)
    """
    return box(val).xor(other).unbox()


def iter(val: OptionalOrMaybe[T]) -> List[T]:
    """
    Wrap ``val`` into a list, if it is not ``None``.

    Allows to use for loops on optional values.
    """
    return box(val).iter()


def zip(
    val: OptionalOrMaybe[T], other: OptionalOrMaybe[U]
) -> Optional[Tuple[T, U]]:
    """
    Return tuple of ``(val, other)`` if neither is ``None``, otherwise return
    ``None``.
    """
    return box(val).zip(other).unbox()


def zip_with(
    val: OptionalOrMaybe[T], other: OptionalOrMaybe[U], fn: Callable[[T, U], R]
) -> Optional[R]:
    """
    Apply function to two optional values, if neither of them is ``None``:

    >>> add = lambda left, right: left + right
    >>> zip_with(1, 2, add)
    3
    >>> zip_with(1, None, add)
    >>> zip_with(None, 2, add)
    """
    return box(val).zip_with(other, fn).unbox()


class Maybe(ABC, Generic[T]):
    @abstractmethod
    def unbox(self) -> Optional[T]:
        """
        Turn ``Maybe[T]`` into ``Optional[T]``.

        >>> Some(1).unbox()
        1
        >>> Some(None).unbox() is None
        True
        """

    @abstractmethod
    def is_some(self) -> bool:
        pass

    def is_none(self) -> bool:
        return not self.is_some()

    @abstractmethod
    def expect(self, msg: str) -> T:
        """
        Ensure that ``val`` is not ``None``, raises a ``ValueError`` using
        ``msg`` otherwise.

        >>> Some(1).expect("My message")
        1
        >>> Nothing.expect("My message")
        Traceback (most recent call last):
            ...
        ValueError: My message
        """

    @abstractmethod
    def do(self, fn: Callable[[T], U]) -> Maybe[T]:
        """
        Apply ``fn`` to ``val`` then return ``val``, if ``val`` is not
        ``None``.

        >>> Some("a").do(print)
        a
        Some('a')
        >>> Nothing.do(print)
        Nothing
        """

    @abstractmethod
    def map(
        self, fn: Callable[Concatenate[T, P], U], *args, **kwargs
    ) -> Maybe[U]:
        """
        Apply ``fn`` to ``val`` if ``val`` is not ``None``.

        >>> Some(1).map(lambda x: x + 1)
        Some(2)
        >>> Nothing.map(lambda x: x + 1)
        Nothing

        Allows to pass additional arguments that are passed to ``fn``:

        >>> Some(10).map(divmod, 3)
        Some((3, 1))
        """

    @abstractmethod
    def map_or(self, fn: Callable[[T], U], default: U) -> U:
        """
        Apply ``fn`` to ``val`` if ``val`` is not ``None`` and return the
        result. In case of ``None`` the provided ``default`` is returned
        instead.

        This is similar to calling ``map`` and ``unwrap_or`` in succession.

        >>> Some(["x"]).map_or(len, 0)
        1
        >>> Nothing.map_or(len, 0)
        0
        """

    @abstractmethod
    def map_or_else(
        self,
        fn: Callable[[T], U],
        factory: Callable[[], U],
    ) -> U:
        """
        Similar to ``map_or``, except that the returned value is lazily
        evaluated.

        This is similar to calling ``map`` and ``unwrap_or_else`` in
        succession.

        >>> Some(1).map_or_else(lambda n: [n], list)
        [1]
        >>> Nothing.map_or_else(lambda n: [n], list)
        []
        """

    @abstractmethod
    def unwrap(self) -> T:
        """
        Assert that the value is not ``None``.

        >>> Some(1).unwrap()
        1
        >>> Nothing.unwrap()
        Traceback (most recent call last):
            ...
        ValueError: Trying to unwrap `None` value.
        """

    @abstractmethod
    def unwrap_or(self, default: T) -> T:
        """
        Get ``val`` if it is not ``None``, or ``default`` otherwise.

        >>> Some(1).unwrap_or(2)
        1
        >>> Nothing.unwrap_or(2)
        2
        """

    @abstractmethod
    def unwrap_or_else(self, fn: Callable[[], T]) -> T:
        """
        Get ``val`` if it is not ``None``, or invoke ``factory`` to get a
        fallback.

        >>> Some([1, 2, 3]).unwrap_or_else(list)
        [1, 2, 3]
        >>> Nothing.unwrap_or_else(list)
        []
        """

    @abstractmethod
    def and_(self, other: OptionalOrMaybe[U]) -> Maybe[U]:
        """
        Like ``a and b`` in Python, except only considering ``None``.

        This is implement identical to ``unwrap_or``, but different with
        respect to types.

        >>> Some(1).and_(2)
        Some(2)
        >>> Some(1).and_(None)
        Nothing
        >>> Nothing.and_(2)
        Nothing
        """

    def __and__(self, other: OptionalOrMaybe[U]) -> Maybe[U]:
        """
        Like ``a and b`` in Python, except only considering ``None``.

        This is implement identical to ``unwrap_or``, but different with
        respect to types.

        >>> Some(1) & 2
        Some(2)
        >>> Some(1) & None
        Nothing
        >>> Nothing & 2
        Nothing
        """
        return self.and_(other)

    @abstractmethod
    def and_then(
        self,
        fn: Callable[Concatenate[T, P], OptionalOrMaybe[U]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Maybe[U]:
        """
        Apply ``fn`` to ``val`` if it is not ``None`` and return the result.

        In contrast to ``map``, ``fn`` always returns an ``Optional``, which is
        consequently flattened.

        >>> Some([42]).and_then(lambda xs: xs[0] if xs else None)
        Some(42)
        >>> Some([]).and_then(lambda xs: xs[0] if xs else None)
        Nothing
        >>> Nothing.and_then(lambda xs: xs[0] if xs else None)
        Nothing
        """

    @abstractmethod
    def or_(self, default: Optional[T]) -> Maybe[T]:
        """
        Like ``a or b`` in Python, except only considering ``None``.

        >>> Some(1).or_(2)
        Some(1)
        >>> Some(1).or_(None)
        Some(1)
        >>> Nothing.or_(2)
        Some(2)
        """

    def __or__(self, default: Optional[T]) -> Maybe[T]:
        """
        Like ``a or b`` in Python, except only considering ``None``.

        >>> Some(1) | 2
        Some(1)
        >>> Some(1) | None
        Some(1)
        >>> Nothing | 2
        Some(2)
        """

        return self.or_(default)

    @abstractmethod
    def or_else(self, factory: Callable[[], Optional[T]]) -> Maybe[T]:
        """
        Like `unwrap_or_else`, except that it returns an optional value.

        >>> Some([42]).or_else(list)
        Some([42])
        >>> Nothing.or_else(list)
        Some([])
        """

    @abstractmethod
    def contains(self, other: U) -> bool:
        """
        Check if ``val`` equals ``other``, always return ``False`` if ``val``
        is ``None``.

        >>> Some(1).contains(1)
        True
        >>> Some(1).contains(2)
        False
        >>> Nothing.contains(3)
        False
        """

    @abstractmethod
    def filter(self, pred: Callable[[T], bool]) -> Maybe[T]:
        """
        Return ``None`` if ``val`` is ``None`` or if ``pred(val)`` does not
        return ``True``, otherwise return ``val``.

        >>> is_even = lambda n: n % 2 == 0
        >>> Some(1).filter(is_even)
        Nothing
        >>> Some(2).filter(is_even)
        Some(2)
        >>> Nothing.filter(is_even)
        Nothing
        """

    @abstractmethod
    def xor(self, other: OptionalOrMaybe[T]) -> Maybe[T]:
        """
        Return either ``val`` or ``other`` if the other is ``None``. Also
        return ``None`` if both are not ``None``.

        >>> xor(1, None)
        1
        >>> xor(None, 2)
        2
        >>> xor(1, 2)
        >>> xor(None, None)
        """

    def __xor__(self, other: OptionalOrMaybe[T]) -> Maybe[T]:
        return self.xor(other)

    @abstractmethod
    def iter(self) -> List[T]:
        """
        Wrap ``val`` into a list, if it is not ``None``.

        Allows to use for loops on optional values.
        """

    def __iter__(self):
        yield from self.iter()

    @abstractmethod
    def zip(self, other: OptionalOrMaybe[U]) -> Maybe[Tuple[T, U]]:
        """
        Abstract zip.
        """

    @abstractmethod
    def zip_with(
        self, other: OptionalOrMaybe[U], fn: Callable[[T, U], R]
    ) -> Maybe[R]:
        """Apply function to two optional values, if neither of them is
        ``None``:

        >>> add = lambda left, right: left + right
        >>> Some(1).zip_with(2, add)
        Some(3)
        >>> Some(1).zip_with(None, add)
        Nothing
        >>> Nothing.zip_with(2, add)
        Nothing
        """

    @abstractmethod
    def flatten(self: "Maybe[OptionalOrMaybe[T]]") -> Maybe[T]:
        """
        Flatten nested optional value.

        Note: This just returns the value, but changes the type from
        ``Optional[Optional[T]]`` to ``Optional[T].``
        """


@dataclass
@final
class Some(Maybe[T]):
    val: T

    def __repr__(self):
        return f"Some({self.val!r})"

    def unbox(self) -> Optional[T]:
        return self.val

    def is_some(self):
        return True

    def unwrap(self) -> T:
        return self.val

    def expect(self, msg: str) -> T:
        return self.val

    def do(self, fn: Callable[[T], U]) -> Maybe[T]:
        fn(self.val)

        return self

    def map(
        self, fn: Callable[Concatenate[T, P], U], *args, **kwargs
    ) -> Maybe[U]:
        return Some(fn(self.val, *args, **kwargs))

    def map_or(self, fn: Callable[[T], U], default: U) -> U:
        return self.map(fn).unwrap()

    def map_or_else(
        self,
        fn: Callable[[T], U],
        factory: Callable[[], U],
    ) -> U:
        return self.map(fn).unwrap()

    def unwrap_or(self, default: T) -> T:
        return self.unwrap()

    def unwrap_or_else(self, fn: Callable[[], T]) -> T:
        return self.unwrap()

    def and_(self, other: OptionalOrMaybe[U]) -> Maybe[U]:
        return box(other)

    def __and__(self, other: OptionalOrMaybe[U]) -> Maybe[U]:
        return self.and_(other)

    def and_then(
        self,
        fn: Callable[Concatenate[T, P], OptionalOrMaybe[U]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Maybe[U]:
        return box(fn(self.val, *args, **kwargs))

    def or_(self, default: Optional[T]) -> Maybe[T]:
        return self

    def __or__(self, default: Optional[T]) -> Maybe[T]:
        return self.or_(default)

    def or_else(self, factory: Callable[[], Optional[T]]) -> Maybe[T]:
        return self

    def contains(self, other: U) -> bool:
        return self.val == other

    def filter(self, pred: Callable[[T], bool]) -> Maybe[T]:
        if pred(self.val):
            return self

        return Nothing

    def xor(self, other: OptionalOrMaybe[T]) -> Maybe[T]:
        other = box(other)

        if other.is_none():
            return self

        return Nothing

    def iter(self) -> List[T]:
        return [self.val]

    def zip(self, other: OptionalOrMaybe[U]) -> Maybe[Tuple[T, U]]:
        other = box(other)
        if other.is_some():
            return Some((self.unwrap(), other.unwrap()))

        return Nothing

    def zip_with(
        self, other: OptionalOrMaybe[U], fn: Callable[[T, U], R]
    ) -> Maybe[R]:
        zipped = self.zip(other)

        if zipped.is_some():
            return box(fn(*zipped.unwrap()))

        return Nothing

    def flatten(self: "Maybe[OptionalOrMaybe[T]]") -> Maybe[T]:
        return box(self.unwrap())


@final
class _Nothing(Maybe[T]):
    def __repr__(self):
        return "Nothing"

    def unbox(self) -> Optional[T]:
        return None

    def is_some(self):
        return False

    def unwrap(self) -> T:
        self.expect("Trying to unwrap `None` value.")

        assert False

    def expect(self, msg: str) -> T:
        raise ValueError(msg)

    def do(self, fn: Callable[[T], U]) -> Maybe[T]:
        return self

    def map(
        self, fn: Callable[Concatenate[T, P], U], *args, **kwargs
    ) -> Maybe[U]:
        return Nothing

    def map_or(self, fn: Callable[[T], U], default: U) -> U:
        return default

    def map_or_else(
        self,
        fn: Callable[[T], U],
        factory: Callable[[], U],
    ) -> U:
        return factory()

    def unwrap_or(self, default: T) -> T:
        return default

    def unwrap_or_else(self, fn: Callable[[], T]) -> T:
        return fn()

    def and_(self, other: OptionalOrMaybe[U]) -> Maybe[U]:
        return Nothing

    def and_then(
        self,
        fn: Callable[Concatenate[T, P], OptionalOrMaybe[U]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Maybe[U]:
        return Nothing

    def or_(self, default: Optional[T]) -> Maybe[T]:
        return box(default)

    def or_else(self, factory: Callable[[], Optional[T]]) -> Maybe[T]:
        return box(factory())

    def contains(self, other: U) -> bool:
        return False

    def filter(self, pred: Callable[[T], bool]) -> Maybe[T]:
        return self

    def xor(self, other: OptionalOrMaybe[T]) -> Maybe[T]:
        return box(other)

    def iter(self) -> List[T]:
        return []

    def zip(self, other: OptionalOrMaybe[U]) -> Maybe[Tuple[T, U]]:
        return Nothing

    def zip_with(
        self, other: OptionalOrMaybe[U], fn: Callable[[T, U], R]
    ) -> Maybe[R]:
        return Nothing

    def flatten(self: "Maybe[OptionalOrMaybe[T]]") -> Maybe[T]:
        return cast(Maybe[T], self)


Nothing: _Nothing[Any] = _Nothing()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import statsmodels.api as sm

from gluonts.core.component import validated
from gluonts.dataset.common import DataEntry
from gluonts.dataset.util import forecast_start
from gluonts.model.forecast import Forecast, SampleForecast
from gluonts.model.predictor import RepresentablePredictor


def seasonality_test(past_ts_data: np.ndarray, season_length: int) -> bool:
    """
    Test the time series for seasonal patterns by performing a 90% auto-
    correlation test.

    For details, see:
    http://www.unic.ac.cy/test/wp-content/uploads/sites/2/2018/09/M4-Competitors-Guide.pdf

    Code based on:
    https://github.com/Mcompetitions/M4-methods/blob/master/Benchmarks%20and%20Evaluation.R
    """
    critical_z_score = 1.645  # corresponds to 90% prediction interval
    if len(past_ts_data) < 3 * season_length:
        return False
    else:
        # calculate auto-correlation for lags up to season_length
        auto_correlations = sm.tsa.stattools.acf(
            past_ts_data, fft=False, nlags=season_length
        )
        auto_correlations[1:] = 2 * auto_correlations[1:] ** 2
        limit = (
            critical_z_score
            / np.sqrt(len(past_ts_data))
            * np.sqrt(np.cumsum(auto_correlations))
        )
        is_seasonal = (
            abs(auto_correlations[season_length]) > limit[season_length]
        )

    return is_seasonal


def naive_2(
    past_ts_data: np.ndarray,
    prediction_length: int,
    season_length: int,
) -> np.ndarray:
    """
    Make seasonality adjusted time series prediction.

    For details, see:
    http://www.unic.ac.cy/test/wp-content/uploads/sites/2/2018/09/M4-Competitors-Guide.pdf

    Code based on:
    https://github.com/Mcompetitions/M4-methods/blob/master/Benchmarks%20and%20Evaluation.R
    """
    assert season_length > 0, "The value of `season_length` should be > 0"

    has_seasonality = False

    if season_length > 1:
        has_seasonality = seasonality_test(past_ts_data, season_length)

    # it has seasonality, then calculate the multiplicative seasonal component
    if has_seasonality:
        # TODO: think about maybe only using past_ts_data[- max
        # (5*season_length, 2*prediction_length):] for speedup
        seasonal_decomposition = sm.tsa.seasonal_decompose(
            x=past_ts_data, period=season_length, model="multiplicative"
        ).seasonal
        seasonality_normed_context = past_ts_data / seasonal_decomposition

        last_period = seasonal_decomposition[-season_length:]
        num_required_periods = (prediction_length // len(last_period)) + 1

        multiplicative_seasonal_component = np.tile(
            last_period, num_required_periods
        )[:prediction_length]
    else:
        seasonality_normed_context = past_ts_data
        multiplicative_seasonal_component = np.ones(
            prediction_length
        )  # i.e. no seasonality component

    # calculate naive forecast: (last value prediction_length times)
    naive_forecast = (
        np.ones(prediction_length) * seasonality_normed_context[-1]
    )

    forecast = np.mean(naive_forecast) * multiplicative_seasonal_component

    return forecast


class Naive2Predictor(RepresentablePredictor):
    """
    Naïve 2 forecaster as described in the M4 Competition Guide.

    See: http://www.unic.ac.cy/test/wp-content/uploads/sites/2/2018/09/M4-Competitors-Guide.pdf.

    The Python analogue implementation to:
    https://github.com/Mcompetitions/M4-methods/blob/master/Benchmarks%20and%20Evaluation.R#L118

    Parameters
    ----------
    prediction_length
        Number of time points to predict
    season_length
        Length of the seasonality pattern of the input data
    """

    @validated()
    def __init__(
        self,
        prediction_length: int,
        season_length: int,
    ) -> None:
        super().__init__(prediction_length=prediction_length)

        assert season_length > 0, "The value of `season_length` should be > 0"

        self.prediction_length = prediction_length
        self.season_length = season_length

    def predict_item(self, item: DataEntry) -> Forecast:
        past_ts_data = item["target"]
        item_id = item.get("item_id", None)
        forecast_start_time = forecast_start(item)

        assert (
            len(past_ts_data) >= 1
        ), "all time series should have at least one data point"

        prediction = naive_2(
            past_ts_data, self.prediction_length, self.season_length
        )

        samples = np.array([prediction])

        return SampleForecast(
            samples=samples,
            start_date=forecast_start_time,
            item_id=item_id,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# type: ignore


import pytest
from itertools import product
from typing import List, Tuple, Union

import mxnet as mx

from gluonts.mx import Tensor
from gluonts.mx.distribution import (
    BetaOutput,
    DeterministicOutput,
    DirichletMultinomialOutput,
    DirichletOutput,
    DistributionOutput,
    GammaOutput,
    GaussianOutput,
    LaplaceOutput,
    LowrankMultivariateGaussianOutput,
    MixtureDistributionOutput,
    MultivariateGaussianOutput,
    NegativeBinomialOutput,
    PiecewiseLinearOutput,
    PoissonOutput,
    StudentTOutput,
    UniformOutput,
)

TEST_CASES = [
    (
        GaussianOutput(),
        mx.nd.random.normal(shape=(3, 4, 5, 6)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        StudentTOutput(),
        mx.nd.random.normal(shape=(3, 4, 5, 6)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        GammaOutput(),
        mx.nd.random.gamma(shape=(3, 4, 5, 6)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        BetaOutput(),
        mx.nd.random.gamma(shape=(3, 4, 5, 6)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        LaplaceOutput(),
        mx.nd.random.normal(shape=(3, 4, 5, 6)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        NegativeBinomialOutput(),
        mx.nd.random.normal(shape=(3, 4, 5, 6)),
        [None],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        UniformOutput(),
        mx.nd.random.normal(shape=(3, 4, 5, 6)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        MixtureDistributionOutput([GaussianOutput(), StudentTOutput()]),
        mx.nd.random.normal(shape=(3, 4, 5, 6)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        PoissonOutput(),
        mx.nd.random.normal(shape=(3, 4, 5, 6)),
        [None],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        DeterministicOutput(42.0),
        mx.nd.random.normal(shape=(3, 4, 5, 6)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
    (
        MultivariateGaussianOutput(dim=5),
        mx.nd.random.normal(shape=(3, 4, 10)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None],
        (3, 4),
        (5,),
    ),
    (
        LowrankMultivariateGaussianOutput(dim=5, rank=4),
        mx.nd.random.normal(shape=(3, 4, 10)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None],
        (3, 4),
        (5,),
    ),
    (
        DirichletOutput(dim=5),
        mx.nd.random.gamma(shape=(3, 4, 5)),
        [None],
        [None],
        (3, 4),
        (5,),
    ),
    (
        DirichletMultinomialOutput(dim=5, n_trials=10),
        mx.nd.random.gamma(shape=(3, 4, 5)),
        [None],
        [None],
        (3, 4),
        (5,),
    ),
]

TEST_CASES_WITHOUT_VARIANCE = [
    (
        MixtureDistributionOutput(
            [
                MultivariateGaussianOutput(dim=5),
                MultivariateGaussianOutput(dim=5),
            ]
        ),
        mx.nd.random.normal(shape=(3, 4, 10)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None],
        (3, 4),
        (5,),
    ),
]

TEST_CASES_WITHOUT_MEAN_NOR_VARIANCE = [
    (
        PiecewiseLinearOutput(num_pieces=3),
        mx.nd.random.normal(shape=(3, 4, 5, 6)),
        [None, mx.nd.ones(shape=(3, 4, 5))],
        [None, mx.nd.ones(shape=(3, 4, 5))],
        (3, 4, 5),
        (),
    ),
]


@pytest.mark.parametrize(
    "distr_out, data, loc, scale, expected_batch_shape, expected_event_shape",
    TEST_CASES
    + TEST_CASES_WITHOUT_VARIANCE
    + TEST_CASES_WITHOUT_MEAN_NOR_VARIANCE,
)
def test_distribution_output_shapes(
    distr_out: DistributionOutput,
    data: Tensor,
    loc: List[Union[None, Tensor]],
    scale: List[Union[None, Tensor]],
    expected_batch_shape: Tuple,
    expected_event_shape: Tuple,
):
    args_proj = distr_out.get_args_proj()
    args_proj.initialize()

    args = args_proj(data)

    assert distr_out.event_shape == expected_event_shape

    for l, s in product(loc, scale):
        distr = distr_out.distribution(args, loc=l, scale=s)

        assert distr.batch_shape == expected_batch_shape
        assert distr.event_shape == expected_event_shape

        x = distr.sample()

        assert x.shape == distr.batch_shape + distr.event_shape

        loss = distr.loss(x)

        assert loss.shape == distr.batch_shape

        x1 = distr.sample(num_samples=1)

        assert x1.shape == (1,) + distr.batch_shape + distr.event_shape

        x3 = distr.sample(num_samples=3)

        assert x3.shape == (3,) + distr.batch_shape + distr.event_shape


@pytest.mark.parametrize(
    "distr_out, data, loc, scale, expected_batch_shape, expected_event_shape",
    TEST_CASES + TEST_CASES_WITHOUT_VARIANCE,
)
def test_distribution_output_mean(
    distr_out: DistributionOutput,
    data: Tensor,
    loc: List[Union[None, Tensor]],
    scale: List[Union[None, Tensor]],
    expected_batch_shape: Tuple,
    expected_event_shape: Tuple,
):
    args_proj = distr_out.get_args_proj()
    args_proj.initialize()
    args = args_proj(data)

    for l, s in product(loc, scale):
        distr = distr_out.distribution(args, loc=l, scale=s)
        assert distr.mean.shape == expected_batch_shape + expected_event_shape


@pytest.mark.parametrize(
    "distr_out, data, loc, scale, expected_batch_shape, expected_event_shape",
    TEST_CASES,
)
def test_distribution_output_variance(
    distr_out: DistributionOutput,
    data: Tensor,
    loc: List[Union[None, Tensor]],
    scale: List[Union[None, Tensor]],
    expected_batch_shape: Tuple,
    expected_event_shape: Tuple,
):
    args_proj = distr_out.get_args_proj()
    args_proj.initialize()
    args = args_proj(data)

    for l, s in product(loc, scale):
        distr = distr_out.distribution(args, loc=l, scale=s)
        assert (
            distr.variance.shape
            == expected_batch_shape + expected_event_shape * 2
        )


@pytest.mark.parametrize("value, model_output_shape", [(42.0, (3, 4, 5))])
def test_deterministic_output(value: float, model_output_shape):
    do = DeterministicOutput(value)
    x = mx.nd.ones(model_output_shape)

    args_proj = do.get_args_proj()
    args_proj.initialize()
    args = args_proj(x)
    distr = do.distribution(args)

    s = distr.sample()

    assert (
        (s == value * mx.nd.ones(shape=model_output_shape[:-1]))
        .asnumpy()
        .all()
    )

    assert (distr.prob(s) == 1.0).asnumpy().all()

    s10 = distr.sample(10)

    assert (
        (s10 == value * mx.nd.ones(shape=(10,) + model_output_shape[:-1]))
        .asnumpy()
        .all()
    )

    assert (distr.prob(s10) == 1.0).asnumpy().all()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional, Union, List

import numpy as np
from mxnet.gluon.nn import HybridBlock

from gluonts.core.component import validated
from gluonts.mx import Tensor

from .distribution import getF


class Bijection:
    """
    A bijective transformation.

    This is defined through the forward transformation (computed by the `f`
    method) and the inverse transformation (`f_inv`).
    """

    @validated()
    def __init__(self):
        pass

    def f(self, x: Tensor) -> Tensor:
        r"""
        Forward transformation x -> y.
        """
        raise NotImplementedError

    def f_inv(self, y: Tensor) -> Tensor:
        r"""
        Inverse transformation y -> x.
        """
        raise NotImplementedError

    def log_abs_det_jac(self, x: Tensor, y: Tensor) -> Tensor:
        r"""
        Receives (x, y) and returns log of the absolute value of the Jacobian
        determinant.

        .. math::
            \log |dy/dx|

        Note that this is the Jacobian determinant of the forward
        transformation x -> y.
        """
        raise NotImplementedError

    def inverse_bijection(self) -> "Bijection":
        r"""
        Returns a Bijection instance that represents the inverse of this
        transformation.
        """
        return InverseBijection(self)

    @property
    def event_dim(self) -> int:
        raise NotImplementedError()

    @property
    def sign(self) -> Union[float, Tensor]:
        """
        Return the sign of the Jacobian's determinant.
        """
        raise NotImplementedError()


class InverseBijection(Bijection):
    """
    The inverse of a given transformation.

    This is a wrapper around bijective transformations, that inverts the
    role of `f` and `f_inv`, and modifies other related methods accordingly.

    Parameters
    ----------
    bijection
        The transformation to invert.
    """

    @validated()
    def __init__(self, bijection: Bijection) -> None:
        super().__init__(self)
        self._bijection = bijection

    def f(self, x: Tensor) -> Tensor:
        return self._bijection.f_inv(x)

    def f_inv(self, y: Tensor) -> Tensor:
        return self._bijection.f(y)

    def log_abs_det_jac(self, x: Tensor, y: Tensor) -> Tensor:
        return -self._bijection.log_abs_det_jac(y, x)

    def inverse_bijection(self) -> Bijection:
        return self._bijection

    @property
    def event_dim(self) -> int:
        return self._bijection.event_dim

    @property
    def sign(self) -> Union[float, Tensor]:
        return self._bijection.sign


class ComposedBijection(Bijection):
    """
    Encapsulates a series of bijections and implements functions associated to
    their composition.
    """

    @validated()
    def __init__(self, bijections: Optional[List[Bijection]] = None) -> None:
        super().__init__(self)
        self._bijections: List[Bijection] = []
        if bijections is not None:
            self.__iadd__(bijections)

    @property
    def event_shape(self):
        return self._bijections[0].event_shape

    @property
    def event_dim(self):
        return self._bijections[0].event_dim

    def f(self, x: Tensor) -> Tensor:
        """
        Computes the forward transform of the composition of bijections.

        Parameters
        ----------
        x
            Input Tensor for the forward transform.
        Returns
        -------
        Tensor
            Transformation of x by the forward composition of bijections
        """
        y = x
        for t in self._bijections:
            y = t.f(y)
        return y

    def f_inv(self, y: Tensor) -> Tensor:
        """
        Computes the inverse transform of a composition of bijections.

        Parameters
        ----------
        y
            Input Tensor for the inverse function

        Returns
        -------
        Tensor
            Transformation of y by the inverse composition of bijections
        """
        x = y
        for t in reversed(self._bijections):
            x = t.f_inv(x)
        return x

    def log_abs_det_jac(self, x: Tensor, y: Tensor) -> Tensor:
        """
        Logarithm of the absolute value of the Jacobian determinant
        corresponding to the composed bijection.

        Parameters
        ----------
        x
            input of the forward transformation or output of the inverse
            transform
        y
            output of the forward transform or input of the inverse transform

        Returns
        -------
        Tensor
            Jacobian evaluated for x as input or y as output
        """
        ladj = 0.0
        for t in reversed(self._bijections):
            x = t.f_inv(y)
            # TODO: eventually change for
            # ladj = ladj + sum_trailing_axes(getF(y), t.log_abs_det_jac(x, y),
            #                                 self.event_dim - t.event_dim)
            ladj = ladj + t.log_abs_det_jac(x, y)
            y = x
        return ladj

    def __getitem__(self, index: int):
        return self._bijections[index]

    def __len__(self):
        return len(self._bijections)

    def __iadd__(self, bijections: List[Bijection]):
        for b in bijections:
            if not isinstance(b, Bijection):
                raise TypeError(
                    f"Object is of type {type(b)}"
                    f" but should inherit from {Bijection}."
                )

            if len(self._bijections) > 0 and b.event_shape != self.event_shape:
                raise RuntimeError(
                    f"Bijection {b} has event_shape of '{b.event_shape}'"
                    f"but should be of '{self.event_shape}'"
                )

            self._bijections.append(b)

        return self

    def __add__(self, bijections: List[Bijection]):
        return ComposedBijection(self._bijections + bijections)


class BijectionHybridBlock(HybridBlock, Bijection):
    """
    Allows a Bijection to have parameters.
    """


class ComposedBijectionHybridBlock(BijectionHybridBlock, ComposedBijection):
    """
    Allows a ComposedBijection object to have parameters.
    """

    @validated()
    def __init__(
        self,
        bij_blocks: Optional[List[Bijection]] = None,
        *args,
        **kwargs,
    ) -> None:
        HybridBlock.__init__(self, *args, **kwargs)
        ComposedBijection.__init__(self, bij_blocks)

    def __iadd__(self, bij_blocks: List[Bijection]):
        for b in bij_blocks:
            self.register_child(b)
        return super().__iadd__(bij_blocks)


class _Exp(Bijection):
    def f(self, x: Tensor) -> Tensor:
        return x.clip(-np.inf, 30).exp()

    def f_inv(self, y: Tensor) -> Tensor:
        return y.clip(1.0e-20, np.inf).log()

    def log_abs_det_jac(self, x: Tensor, y: Tensor) -> Tensor:
        return y.clip(1.0e-20, np.inf).log()

    @property
    def event_dim(self) -> int:
        return 0

    @property
    def sign(self) -> float:
        return 1.0


class _Log(Bijection):
    def f(self, x: Tensor) -> Tensor:
        return x.clip(1.0e-20, np.inf).log()

    def f_inv(self, y: Tensor) -> Tensor:
        return y.clip(-np.inf, 30).exp()

    def log_abs_det_jac(self, x: Tensor, y: Tensor) -> Tensor:
        return -y

    @property
    def event_dim(self) -> int:
        return 0

    @property
    def sign(self) -> float:
        return 1.0


class _Softrelu(Bijection):
    def _log_expm1(self, F, y: Tensor) -> Tensor:
        r"""
        A numerically stable computation of :math:`x = \log(e^y - 1)`
        """
        thresh = F.zeros_like(y) + 20.0
        x = F.where(F.broadcast_greater(y, thresh), y, F.log(F.expm1(y)))
        return x

    def f(self, x: Tensor) -> Tensor:
        F = getF(x)
        return F.Activation(x.clip(-100.0, np.inf), act_type="softrelu")

    def f_inv(self, y: Tensor) -> Tensor:
        F = getF(y)
        return self._log_expm1(F, y)

    def log_abs_det_jac(self, x: Tensor, y: Tensor) -> Tensor:
        F = getF(y)
        return self._log_expm1(F, y) - y

    @property
    def event_dim(self) -> int:
        return 0

    @property
    def sign(self) -> float:
        return 1.0


class AffineTransformation(Bijection):
    """
    An affine transformation consisting of a scaling and a translation.

    If translation is specified `loc`, and the scaling by `scale`, then
    this transformation computes `y = scale * x + loc`, where all operations
    are element-wise.

    Parameters
    ----------
    loc
        Translation parameter. If unspecified or `None`, this will be zero.
    scale
        Scaling parameter. If unspecified or `None`, this will be one.
    """

    @validated()
    def __init__(
        self, loc: Optional[Tensor] = None, scale: Optional[Tensor] = None
    ) -> None:
        super().__init__(self)
        self.loc = loc
        self.scale = scale

    def f(self, x: Tensor) -> Tensor:
        F = getF(x)
        if self.scale is not None:
            x = F.broadcast_mul(x, self.scale)
        if self.loc is not None:
            x = F.broadcast_add(x, self.loc)
        return x

    def f_inv(self, y: Tensor) -> Tensor:
        F = getF(y)
        if self.loc is not None:
            y = F.broadcast_sub(y, self.loc)
        if self.scale is not None:
            y = F.broadcast_div(y, self.scale)
        return y

    def log_abs_det_jac(self, x: Tensor, y: Tensor) -> Tensor:
        if self.scale is not None:
            return self.scale.broadcast_like(x).abs().log()
        F = getF(x)
        return F.zeros_like(x)

    @property
    def sign(self) -> Union[float, Tensor]:
        return 1.0 if self.scale is None else self.scale.sign()

    @property
    def event_dim(self) -> int:
        return 0


exp = _Exp()
log = _Log()
softrelu = _Softrelu()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging
from typing import Dict, List, Optional, Tuple

import torch
import torch.nn as nn

from gluonts.core.component import validated
from gluonts.model import Input, InputSpec
from gluonts.torch.distributions import Output
from gluonts.torch.distributions.quantile_output import QuantileOutput
from gluonts.torch.scaler import StdScaler
from gluonts.torch.util import weighted_average

from .layers import (
    FeatureEmbedder,
    FeatureProjector,
    GatedResidualNetwork,
    TemporalFusionDecoder,
    TemporalFusionEncoder,
    VariableSelectionNetwork,
)

logger = logging.getLogger(__name__)


class TemporalFusionTransformerModel(nn.Module):
    """
    Temporal Fusion Transformer neural network.

    Partially based on the implementation in github.com/kashif/pytorch-transformer-ts.

    Inputs feat_static_real, feat_static_cat and feat_dynamic_real are mandatory.
    Inputs feat_dynamic_cat, past_feat_dynamic_real and past_feat_dynamic_cat are optional.
    """

    @validated()
    def __init__(
        self,
        context_length: int,
        prediction_length: int,
        d_feat_static_real: Optional[List[int]] = None,  #  Defaults to [1]
        c_feat_static_cat: Optional[List[int]] = None,  #  Defaults to [1]
        d_feat_dynamic_real: Optional[List[int]] = None,  # Defaults to [1]
        c_feat_dynamic_cat: Optional[List[int]] = None,  # Defaults to []
        d_past_feat_dynamic_real: Optional[List[int]] = None,  # Defaults to []
        c_past_feat_dynamic_cat: Optional[List[int]] = None,  # Defaults to []
        num_heads: int = 4,
        d_hidden: int = 32,
        d_var: int = 32,
        dropout_rate: float = 0.1,
        distr_output: Optional[Output] = None,
    ):
        super().__init__()

        self.context_length = context_length
        self.prediction_length = prediction_length
        self.num_heads = num_heads
        self.d_hidden = d_hidden
        self.d_var = d_var
        self.dropout_rate = dropout_rate

        if distr_output is None:
            distr_output = QuantileOutput(
                [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
            )
        self.distr_output = distr_output
        self.args_proj = self.distr_output.get_args_proj(
            in_features=self.d_hidden
        )

        self.d_feat_static_real = d_feat_static_real or [1]
        self.d_feat_dynamic_real = d_feat_dynamic_real or [1]
        self.d_past_feat_dynamic_real = d_past_feat_dynamic_real or []
        self.c_feat_static_cat = c_feat_static_cat or [1]
        self.c_feat_dynamic_cat = c_feat_dynamic_cat or []
        self.c_past_feat_dynamic_cat = c_past_feat_dynamic_cat or []

        self.num_feat_static = len(self.d_feat_static_real) + len(
            self.c_feat_static_cat
        )
        self.num_feat_dynamic = len(self.d_feat_dynamic_real) + len(
            self.c_feat_dynamic_cat
        )
        self.num_past_feat_dynamic = len(self.d_past_feat_dynamic_real) + len(
            self.c_past_feat_dynamic_cat
        )

        self.scaler = StdScaler(dim=1, keepdim=True)

        self.target_proj = nn.Linear(in_features=1, out_features=self.d_var)
        # Past-only dynamic features
        if self.d_past_feat_dynamic_real:
            self.past_feat_dynamic_proj: Optional[FeatureProjector] = (
                FeatureProjector(
                    feature_dims=self.d_past_feat_dynamic_real,
                    embedding_dims=[self.d_var]
                    * len(self.d_past_feat_dynamic_real),
                )
            )
        else:
            self.past_feat_dynamic_proj = None

        if self.c_past_feat_dynamic_cat:
            self.past_feat_dynamic_embed: Optional[FeatureEmbedder] = (
                FeatureEmbedder(
                    cardinalities=self.c_past_feat_dynamic_cat,
                    embedding_dims=[self.d_var]
                    * len(self.c_past_feat_dynamic_cat),
                )
            )
        else:
            self.past_feat_dynamic_embed = None

        # Known dynamic features
        if self.d_feat_dynamic_real:
            self.feat_dynamic_proj: Optional[FeatureProjector] = (
                FeatureProjector(
                    feature_dims=self.d_feat_dynamic_real,
                    embedding_dims=[self.d_var]
                    * len(self.d_feat_dynamic_real),
                )
            )
        else:
            self.feat_dynamic_proj = None

        if self.c_feat_dynamic_cat:
            self.feat_dynamic_embed: Optional[FeatureEmbedder] = (
                FeatureEmbedder(
                    cardinalities=self.c_feat_dynamic_cat,
                    embedding_dims=[self.d_var] * len(self.c_feat_dynamic_cat),
                )
            )
        else:
            self.feat_dynamic_embed = None

        # Static features
        if self.d_feat_static_real:
            self.feat_static_proj: Optional[FeatureProjector] = (
                FeatureProjector(
                    feature_dims=self.d_feat_static_real,
                    embedding_dims=[self.d_var] * len(self.d_feat_static_real),
                )
            )
        else:
            self.feat_static_proj = None

        if self.c_feat_static_cat:
            self.feat_static_embed: Optional[FeatureEmbedder] = (
                FeatureEmbedder(
                    cardinalities=self.c_feat_static_cat,
                    embedding_dims=[self.d_var] * len(self.c_feat_static_cat),
                )
            )
        else:
            self.feat_static_embed = None

        self.static_selector = VariableSelectionNetwork(
            d_hidden=self.d_var,
            num_vars=self.num_feat_static,
            dropout=self.dropout_rate,
        )
        self.ctx_selector = VariableSelectionNetwork(
            d_hidden=self.d_var,
            num_vars=self.num_past_feat_dynamic + self.num_feat_dynamic + 1,
            add_static=True,
            dropout=self.dropout_rate,
        )
        self.tgt_selector = VariableSelectionNetwork(
            d_hidden=self.d_var,
            num_vars=self.num_feat_dynamic,
            add_static=True,
            dropout=self.dropout_rate,
        )
        self.selection = GatedResidualNetwork(
            d_hidden=self.d_var,
            dropout=self.dropout_rate,
        )
        self.enrichment = GatedResidualNetwork(
            d_hidden=self.d_var,
            dropout=self.dropout_rate,
        )
        self.state_h = GatedResidualNetwork(
            d_hidden=self.d_var,
            d_output=self.d_hidden,
            dropout=self.dropout_rate,
        )
        self.state_c = GatedResidualNetwork(
            d_hidden=self.d_var,
            d_output=self.d_hidden,
            dropout=self.dropout_rate,
        )
        self.temporal_encoder = TemporalFusionEncoder(
            d_input=self.d_var,
            d_hidden=self.d_hidden,
        )
        self.temporal_decoder = TemporalFusionDecoder(
            context_length=self.context_length,
            prediction_length=self.prediction_length,
            d_hidden=self.d_hidden,
            d_var=self.d_var,
            num_heads=self.num_heads,
            dropout=self.dropout_rate,
        )

    def describe_inputs(self, batch_size=1) -> InputSpec:
        return InputSpec(
            {
                "past_target": Input(
                    shape=(batch_size, self.context_length), dtype=torch.float
                ),
                "past_observed_values": Input(
                    shape=(batch_size, self.context_length), dtype=torch.float
                ),
                "feat_static_real": Input(
                    shape=(batch_size, sum(self.d_feat_static_real)),
                    dtype=torch.float,
                ),
                "feat_static_cat": Input(
                    shape=(batch_size, len(self.c_feat_static_cat)),
                    dtype=torch.long,
                ),
                "feat_dynamic_real": Input(
                    shape=(
                        batch_size,
                        self.context_length + self.prediction_length,
                        sum(self.d_feat_dynamic_real),
                    ),
                    dtype=torch.float,
                ),
                "feat_dynamic_cat": Input(
                    shape=(
                        batch_size,
                        self.context_length + self.prediction_length,
                        len(self.c_feat_dynamic_cat),
                    ),
                    dtype=torch.long,
                ),
                "past_feat_dynamic_real": Input(
                    shape=(
                        batch_size,
                        self.context_length,
                        sum(self.d_past_feat_dynamic_real),
                    ),
                    dtype=torch.float,
                ),
                "past_feat_dynamic_cat": Input(
                    shape=(
                        batch_size,
                        self.context_length,
                        len(self.c_past_feat_dynamic_cat),
                    ),
                    dtype=torch.long,
                ),
            },
            torch.zeros,
        )

    def input_types(self) -> Dict[str, torch.dtype]:
        return {
            "past_target": torch.float,
            "past_observed_values": torch.float,
            "feat_static_real": torch.float,
            "feat_static_cat": torch.long,
            "feat_dynamic_real": torch.float,
            "feat_dynamic_cat": torch.long,
            "past_feat_dynamic_real": torch.float,
            "past_feat_dynamic_cat": torch.long,
        }

    def _preprocess(
        self,
        past_target: torch.Tensor,  # [N, T]
        past_observed_values: torch.Tensor,  # [N, T]
        feat_static_real: Optional[torch.Tensor],  # [N, D_sr]
        feat_static_cat: Optional[torch.Tensor],  # [N, D_sc]
        feat_dynamic_real: Optional[torch.Tensor],  # [N, T + H, D_dr]
        feat_dynamic_cat: Optional[torch.Tensor],  # [N, T + H, D_dc]
        past_feat_dynamic_real: Optional[torch.Tensor],  # [N, T, D_pr]
        past_feat_dynamic_cat: Optional[torch.Tensor],  # [N, T, D_pc]
    ) -> Tuple[
        List[torch.Tensor],
        List[torch.Tensor],
        List[torch.Tensor],
        torch.Tensor,
        torch.Tensor,
    ]:
        past_target, loc, scale = self.scaler(
            data=past_target, weights=past_observed_values
        )

        past_covariates = [self.target_proj(past_target.unsqueeze(-1))]
        future_covariates = []
        static_covariates = []
        if self.past_feat_dynamic_proj is not None:
            projs = self.past_feat_dynamic_proj(past_feat_dynamic_real)
            past_covariates.extend(projs)
        if self.past_feat_dynamic_embed is not None:
            embs = self.past_feat_dynamic_embed(past_feat_dynamic_cat)
            past_covariates.extend(embs)

        if self.feat_dynamic_proj is not None:
            projs = self.feat_dynamic_proj(feat_dynamic_real)
            for proj in projs:
                ctx_proj = proj[..., : self.context_length, :]
                tgt_proj = proj[..., self.context_length :, :]
                past_covariates.append(ctx_proj)
                future_covariates.append(tgt_proj)
        if self.feat_dynamic_embed is not None:
            embs = self.feat_dynamic_embed(feat_dynamic_cat)
            for emb in embs:
                ctx_emb = emb[..., : self.context_length, :]
                tgt_emb = emb[..., self.context_length :, :]
                past_covariates.append(ctx_emb)
                future_covariates.append(tgt_emb)

        if self.feat_static_proj is not None:
            projs = self.feat_static_proj(feat_static_real)
            static_covariates.extend(projs)
        if self.feat_static_embed is not None:
            embs = self.feat_static_embed(feat_static_cat)
            static_covariates.extend(embs)

        return (
            past_covariates,
            future_covariates,
            static_covariates,
            loc,
            scale,
        )

    def forward(
        self,
        past_target: torch.Tensor,  # [N, T]
        past_observed_values: torch.Tensor,  # [N, T]
        feat_static_real: Optional[torch.Tensor],  # [N, D_sr]
        feat_static_cat: Optional[torch.Tensor],  # [N, D_sc]
        feat_dynamic_real: Optional[torch.Tensor],  # [N, T + H, D_dr]
        feat_dynamic_cat: Optional[torch.Tensor] = None,  # [N, T + H, D_dc]
        past_feat_dynamic_real: Optional[torch.Tensor] = None,  # [N, T, D_pr]
        past_feat_dynamic_cat: Optional[torch.Tensor] = None,  # [N, T, D_pc]
    ) -> Tuple[Tuple[torch.Tensor, ...], torch.Tensor, torch.Tensor]:
        (
            past_covariates,  # [[N, T, d_var], ...]
            future_covariates,  # [[N, H, d_var], ...]
            static_covariates,  # [[N, d_var], ...]
            loc,  # [N, 1]
            scale,  # [N, 1]
        ) = self._preprocess(
            past_target=past_target,
            past_observed_values=past_observed_values,
            feat_static_real=feat_static_real,
            feat_static_cat=feat_static_cat,
            feat_dynamic_real=feat_dynamic_real,
            feat_dynamic_cat=feat_dynamic_cat,
            past_feat_dynamic_real=past_feat_dynamic_real,
            past_feat_dynamic_cat=past_feat_dynamic_cat,
        )

        static_var, _ = self.static_selector(static_covariates)  # [N, d_var]
        c_selection = self.selection(static_var).unsqueeze(1)  # [N, 1, d_var]
        c_enrichment = self.enrichment(static_var).unsqueeze(1)
        c_h = self.state_h(static_var)  # [N, self.d_hidden]
        c_c = self.state_c(static_var)  # [N, self.d_hidden]
        states = [c_h.unsqueeze(0), c_c.unsqueeze(0)]

        ctx_input, _ = self.ctx_selector(
            past_covariates, c_selection
        )  # [N, T, d_var]
        tgt_input, _ = self.tgt_selector(
            future_covariates, c_selection
        )  # [N, H, d_var]

        encoding = self.temporal_encoder(
            ctx_input, tgt_input, states
        )  # [N, T + H, d_hidden]
        decoding = self.temporal_decoder(
            encoding, c_enrichment, past_observed_values
        )  # [N, H, d_hidden]
        distr_args = self.args_proj(decoding)
        return distr_args, loc, scale

    def loss(
        self,
        past_target: torch.Tensor,  # [N, T]
        past_observed_values: torch.Tensor,  # [N, T]
        future_target: torch.Tensor,  # [N, H]
        future_observed_values: torch.Tensor,  # [N, H]
        feat_static_real: torch.Tensor,  # [N, D_sr]
        feat_static_cat: torch.Tensor,  # [N, D_sc]
        feat_dynamic_real: torch.Tensor,  # [N, T + H, D_dr]
        feat_dynamic_cat: Optional[torch.Tensor] = None,  # [N, T + H, D_dc]
        past_feat_dynamic_real: Optional[torch.Tensor] = None,  # [N, T, D_pr]
        past_feat_dynamic_cat: Optional[torch.Tensor] = None,  # [N, T, D_pc]
    ) -> torch.Tensor:
        distr_args, loc, scale = self(
            past_target=past_target,
            past_observed_values=past_observed_values,
            feat_static_real=feat_static_real,
            feat_static_cat=feat_static_cat,
            feat_dynamic_real=feat_dynamic_real,
            feat_dynamic_cat=feat_dynamic_cat,
            past_feat_dynamic_real=past_feat_dynamic_real,
            past_feat_dynamic_cat=past_feat_dynamic_cat,
        )  # [N, Q, T]
        loss = self.distr_output.loss(
            target=future_target,
            distr_args=distr_args,
            loc=loc,
            scale=scale,
        )
        return weighted_average(loss, weights=future_observed_values, dim=-1)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Tuple

import mxnet as mx
import numpy as np

from gluonts.core.component import validated
from gluonts.mx import Tensor

from .binning_helpers import (
    bin_edges_from_bin_centers,
    ensure_binning_monotonicity,
)
from .representation import Representation


class LocalAbsoluteBinning(Representation):
    """
    A class representing a local absolute binning approach. This binning
    estimates a binning for every single time series on a local level and
    therefore implicitly acts as a scaling mechanism.

    Parameters
    ----------
    num_bins
        The number of discrete bins/buckets that we want values to be mapped
        to. (default: 1024)
    is_quantile
        Whether the binning is quantile or linear. Quantile binning allocated
        bins based on the cumulative distribution function, while linear
        binning allocates evenly spaced bins.(default: True, i.e. quantile
        binning)
    """

    @validated()
    def __init__(
        self, num_bins: int = 1024, is_quantile: bool = True, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.num_bins = num_bins
        self.is_quantile = is_quantile

    def hybrid_forward(
        self,
        F,
        data: Tensor,
        observed_indicator: Tensor,
        scale: Optional[Tensor],
        rep_params: List[Tensor],
        **kwargs,
    ) -> Tuple[Tensor, Tensor, List[Tensor]]:
        data_np = data.asnumpy()
        observed_indicator_np = observed_indicator.astype("int32").asnumpy()

        if scale is None:
            # Even though local binning implicitly scales the data, we still
            # return the scale as an input to the model.
            scale = F.expand_dims(
                F.sum(data * observed_indicator, axis=-1)
                / F.sum(observed_indicator, axis=-1),
                -1,
            )

            bin_centers_hyb = np.ones((len(data), self.num_bins)) * (-1)
            bin_edges_hyb = np.ones((len(data), self.num_bins + 1)) * (-1)

            # Every time series needs to be binned individually
            for i in range(len(data_np)):
                # Identify observed data points.
                data_loc = data_np[i]
                observed_indicator_loc = observed_indicator_np[i]
                data_obs_loc = data_loc[observed_indicator_loc == 1]

                if data_obs_loc.size > 0:
                    # Calculate time series specific bin centers and edges.
                    if self.is_quantile:
                        bin_centers_loc = np.quantile(
                            data_obs_loc, np.linspace(0, 1, self.num_bins)
                        )
                    else:
                        bin_centers_loc = np.linspace(
                            np.min(data_obs_loc),
                            np.max(data_obs_loc),
                            self.num_bins,
                        )
                    bin_centers_hyb[i] = ensure_binning_monotonicity(
                        bin_centers_loc
                    )
                    bin_edges_hyb[i] = bin_edges_from_bin_centers(
                        bin_centers_hyb[i]
                    )

                    # Bin the time series.
                    data_obs_loc_binned = np.digitize(
                        data_obs_loc, bins=bin_edges_hyb[i], right=False
                    )
                else:
                    data_obs_loc_binned = []

                # Write the binned time series back into the data array.
                data_loc[observed_indicator_loc == 1] = data_obs_loc_binned
                data_np[i] = data_loc

        else:
            bin_centers_hyb = rep_params[0].asnumpy()
            bin_edges_hyb = rep_params[1].asnumpy()

            bin_edges_hyb = np.repeat(
                bin_edges_hyb,
                len(data_np) // len(bin_edges_hyb),
                axis=0,
            )
            bin_centers_hyb = np.repeat(
                bin_centers_hyb,
                len(data_np) // len(bin_centers_hyb),
                axis=0,
            )

            for i in range(len(data_np)):
                data_loc = data_np[i]
                observed_indicator_loc = observed_indicator_np[i]
                data_obs_loc = data_loc[observed_indicator_loc == 1]

                # Bin the time series based on previously computed bin edges.
                data_obs_loc_binned = np.digitize(
                    data_obs_loc, bins=bin_edges_hyb[i], right=False
                )

                data_loc[observed_indicator_loc == 1] = data_obs_loc_binned
                data_np[i] = data_loc

        bin_centers_hyb = F.array(bin_centers_hyb)
        bin_edges_hyb = F.array(bin_edges_hyb)

        data = mx.nd.array(data_np)

        return data, scale, [bin_centers_hyb, bin_edges_hyb]

    def post_transform(
        self, F, samples: Tensor, scale: Tensor, rep_params: List[Tensor]
    ) -> Tensor:
        bin_centers_hyb = rep_params[0]

        transf_samples = F.one_hot(F.squeeze(samples), self.num_bins)

        # Pick corresponding bin centers for all samples
        transf_samples = F.sum(
            bin_centers_hyb * transf_samples, axis=1
        ).expand_dims(-1)

        return transf_samples



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List

import pandas as pd
import numpy as np
import pytest

from gluonts.dataset.common import Dataset, ListDataset
from gluonts.dataset.util import to_pandas
from gluonts.transform.feature import AddTimeFeatures
from gluonts.time_feature import (
    TimeFeature,
    month_of_year,
    time_features_from_frequency_str,
)


def compute_time_features(
    entry: Dict,
    time_features: List[TimeFeature],
    pred_length: int = 0,
    dtype=np.float32,
):
    assert pred_length >= 0
    index = to_pandas(entry, freq=entry["start"].freq).index

    if pred_length > 0:
        index = index.union(
            pd.period_range(
                index[-1] + 1, index[-1] + pred_length, freq=index.freq
            )
        )

    feature_arrays = [feat(index) for feat in time_features]
    return np.vstack(feature_arrays).astype(dtype)


@pytest.mark.parametrize(
    "dataset, time_features",
    [
        (
            ListDataset(
                data_iter=[
                    {"start": "2021-01-01 00:00:06", "target": [1.0] * 50},
                    {"start": "2021-01-12 00:45:17", "target": [1.0] * 50},
                    {"start": "2021-02-18 12:00:28", "target": [1.0] * 50},
                    {"start": "2021-05-27 07:10:39", "target": [1.0] * 50},
                ],
                freq=freq_str,
            ),
            time_features_from_frequency_str(freq_str) + [month_of_year],
        )
        for freq_str in [
            "2M",
            "3W",
            "3W-MON",
            "3W-TUE",
            "3W-WED",
            "3W-THU",
            "3W-FRI",
            "3W-SAT",
            "2D",
            "5H",
        ]
    ],
)
@pytest.mark.parametrize("pred_length", [0, 1, 2, 10])
def test_AddTimeFeatures_correctness(
    dataset: Dataset, time_features: List[TimeFeature], pred_length: int
):
    transform = AddTimeFeatures(
        start_field="start",
        target_field="target",
        output_field="features",
        time_features=time_features,
        pred_length=pred_length,
    )
    for entry, transformed_entry in zip(
        dataset, transform(dataset, is_train=(pred_length == 0))
    ):
        expected_features = compute_time_features(
            entry, time_features, pred_length
        )
        assert np.allclose(expected_features, transformed_entry["features"])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, Optional, Union, cast

import mxnet as mx
import numpy as np
import pandas as pd
from pandas import to_timedelta

from gluonts.model.forecast import Forecast


class PointProcessSampleForecast(Forecast):
    """
    Sample forecast object used for temporal point process inference. Differs
    from standard forecast objects as it does not implement fixed length
    samples. Each sample has a variable length, that is kept in a separate
    :code:`valid_length` attribute.

    Importantly, PointProcessSampleForecast does not implement some
    methods (such as :code:`quantile` or :code:`plot`) that are available
    in discrete time forecasts.

    Parameters
    ----------
    samples
        A multidimensional array of samples, of shape
        (number_of_samples, max_pred_length, target_dim). The target_dim is
        equal to 2, where the first dimension contains the inter-arrival times
        and the second - categorical marks.
    valid_length
        An array of integers denoting the valid lengths of each sample
        in :code:`samples`. That is, :code:`valid_length[0] == 2` implies
        that only the first two entries of :code:`samples[0, ...]` are
        valid "points".
    start_date
        Starting Timestamp of the sample
    freq
        The time unit of interarrival times
    prediction_interval_length
        The length of the prediction interval for which samples were drawn.
    item_id
        Item ID, if available.
    info
        Optional dictionary of additional information.

    """

    prediction_interval_length: float

    # TODO: not used
    prediction_length = cast(int, None)
    mean = cast(np.ndarray, None)
    _index = None

    def __init__(
        self,
        samples: Union[mx.nd.NDArray, np.ndarray],
        valid_length: Union[mx.nd.NDArray, np.ndarray],
        start_date: pd.Timestamp,
        freq: str,
        prediction_interval_length: float,
        item_id: Optional[str] = None,
        info: Optional[Dict] = None,
    ) -> None:
        assert isinstance(
            samples, (np.ndarray, mx.nd.NDArray)
        ), "samples should be either a numpy or an mxnet array"
        assert samples.ndim == 2 or samples.ndim == 3, (
            "samples should be a 2-dimensional or 3-dimensional array."
            f" Dimensions found: {samples.ndim}"
        )

        assert isinstance(
            valid_length, (np.ndarray, mx.nd.NDArray)
        ), "samples should be either a numpy or an mxnet array"
        assert (
            valid_length.ndim == 1
        ), "valid_length should be a 1-dimensional array"
        assert (
            valid_length.shape[0] == samples.shape[0]
        ), "valid_length and samples should have compatible dimensions"

        self.samples, self.valid_length = (
            x if isinstance(x, np.ndarray) else x.asnumpy()
            for x in (samples, valid_length)
        )

        self._dim = samples.ndim
        self.item_id = item_id
        self.info = info

        assert isinstance(
            start_date, pd.Timestamp
        ), "start_date should be a pandas Timestamp object"
        self.start_date = start_date

        assert isinstance(freq, str), "freq should be a string"
        self.interarrival_freq = freq

        assert (
            prediction_interval_length > 0
        ), "prediction_interval_length must be greater than 0"
        self.prediction_interval_length = prediction_interval_length

        self.end_date = (
            start_date + to_timedelta(1, freq) * prediction_interval_length
        )

    def dim(self) -> int:
        return self._dim

    @property
    def freq(self):
        return self.interarrival_freq

    @property
    def index(self) -> pd.PeriodIndex:
        raise AttributeError(
            "Period index not defined for point process samples"
        )

    def __repr__(self):
        return ", ".join(
            [
                f"PointProcessSampleForecast({self.samples!r})",
                f"{self.valid_length!r}",
                f"{self.start_date!r}",
                f"{self.end_date!r}",
                f"{self.freq!r}",
                f"item_id={self.item_id!r}",
                f"info={self.info!r})",
            ]
        )

    def quantile(self, q: Union[float, str]) -> np.ndarray:
        raise NotImplementedError(
            "Quantile function is not defined for point process samples"
        )

    def plot(self, **kwargs):
        raise NotImplementedError(
            "Plotting not implemented for point process samples"
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import numpy as np
import pandas as pd
import pytest

from gluonts.model.seasonal_agg import SeasonalAggregatePredictor
from gluonts.transform.feature import LastValueImputation, LeavesMissingValues

FREQ = "D"
START_DATE = "2023"


def get_prediction(
    target,
    prediction_length=1,
    season_length=1,
    num_seasons=1,
    agg_fun=np.nanmean,
    imputation_method=LastValueImputation(),
):
    pred = SeasonalAggregatePredictor(
        prediction_length=prediction_length,
        season_length=season_length,
        num_seasons=num_seasons,
        agg_fun=agg_fun,
        imputation_method=imputation_method,
    )
    item = {
        "target": np.asarray(target),
        "start": pd.Period(START_DATE, freq=FREQ),
    }
    forecast = pred.predict_item(item)

    return forecast


@pytest.mark.parametrize(
    "data, expected_output, prediction_length, season_length, "
    "num_seasons, agg_fun, imputation_method",
    [
        # same as seasonal naive
        ([1, 1, 1], [1], 1, 1, 1, np.nanmean, LastValueImputation()),
        (
            [1, 10, 2, 20],
            [1.5, 15],
            2,
            2,
            2,
            np.nanmean,
            LastValueImputation(),
        ),
        # check predictions repeat seasonally
        (
            [1, 10, 2, 20],
            [1.5, 15, 1.5, 15],
            4,
            2,
            2,
            np.nanmean,
            LastValueImputation(),
        ),
        (
            [1, 10, 2, 20],
            [1.5, 15, 1.5],
            3,
            2,
            2,
            np.nanmean,
            LastValueImputation(),
        ),
        # check `nanmedian`
        (
            [1, 10, 2, 20, 3, 30],
            [2, 20, 2, 20],
            4,
            2,
            3,
            np.nanmedian,
            LastValueImputation(),
        ),
        (
            [1, 10, 2, 20, 3, 30],
            [2, 20, 2],
            3,
            2,
            3,
            np.nanmedian,
            LastValueImputation(),
        ),
        # check `nanmax`
        (
            [1, 10, 2, 20, 3, 30],
            [3, 30, 3, 30],
            4,
            2,
            3,
            np.nanmax,
            LastValueImputation(),
        ),
        # check `nanmin`
        (
            [1, 10, 2, 20, 3, 30],
            [1, 10, 1, 10],
            4,
            2,
            3,
            np.nanmin,
            LastValueImputation(),
        ),
        # data is shorter than season length
        ([1, 2, 3], [2], 1, 4, 1, np.nanmean, LastValueImputation()),
        ([10, 1, 100], [10], 1, 4, 1, np.nanmedian, LastValueImputation()),
        ([10, 1, 100], [100], 1, 4, 1, np.nanmax, LastValueImputation()),
        ([10, 1, 100], [1], 1, 4, 1, np.nanmin, LastValueImputation()),
        # data not available for all seasons
        ([1, 2, 3, 4, 5], [3] * 4, 4, 4, 2, np.nanmean, LastValueImputation()),
        (
            [10, 20, 40, 50, 21],
            [21] * 4,
            4,
            4,
            2,
            np.nanmedian,
            LastValueImputation(),
        ),
        (
            [10, 20, 40, 50, 21],
            [50] * 4,
            4,
            4,
            2,
            np.nanmax,
            LastValueImputation(),
        ),
        (
            [10, 20, 40, 50, 21],
            [10] * 4,
            4,
            4,
            2,
            np.nanmin,
            LastValueImputation(),
        ),
        # missing values with imputation
        ([np.nan], [0], 1, 1, 2, np.nanmean, LastValueImputation()),
        ([np.nan], [0], 1, 1, 2, np.nanmedian, LastValueImputation()),
        ([1, 4, np.nan], [3], 1, 3, 2, np.nanmean, LastValueImputation()),
        ([1, 4, np.nan], [4], 1, 3, 2, np.nanmedian, LastValueImputation()),
        (
            [1, 10, np.nan, 1, 10, np.nan],
            [1, 10, 10],
            3,
            3,
            2,
            np.nanmean,
            LastValueImputation(),
        ),
        (
            [1, 10, np.nan, 1, 10, np.nan],
            [1, 10, 10],
            3,
            3,
            2,
            np.nanmedian,
            LastValueImputation(),
        ),
        (
            [1, 10, np.nan, 1, 10, np.nan],
            [1, 10, 10, 1, 10],
            5,
            3,
            2,
            np.nanmax,
            LastValueImputation(),
        ),
        (
            [1, 10, np.nan, 1, 10, np.nan],
            [1, 10, 10, 1, 10],
            5,
            3,
            2,
            np.nanmin,
            LastValueImputation(),
        ),
        # missing values without imputation
        ([1, 3, np.nan], [np.nan], 1, 1, 1, np.nanmean, LeavesMissingValues()),
        (
            [1, 3, np.nan],
            [np.nan],
            1,
            1,
            1,
            np.nanmedian,
            LeavesMissingValues(),
        ),
        ([1, 3, np.nan], [np.nan], 1, 1, 1, np.nanmax, LeavesMissingValues()),
        ([1, 3, np.nan], [np.nan], 1, 1, 1, np.nanmin, LeavesMissingValues()),
        (
            [1, 3, np.nan],
            [np.nan] * 2,
            2,
            1,
            1,
            np.nanmean,
            LeavesMissingValues(),
        ),
        (
            [1, 3, np.nan],
            [np.nan] * 2,
            2,
            1,
            1,
            np.nanmedian,
            LeavesMissingValues(),
        ),
        (
            [1, 3, np.nan],
            [np.nan] * 2,
            2,
            1,
            1,
            np.nanmax,
            LeavesMissingValues(),
        ),
        (
            [1, 3, np.nan],
            [np.nan] * 2,
            2,
            1,
            1,
            np.nanmin,
            LeavesMissingValues(),
        ),
        ([1, 3, np.nan], [3], 1, 1, 2, np.nanmean, LeavesMissingValues()),
        ([1, 3, np.nan], [3], 1, 1, 2, np.nanmedian, LeavesMissingValues()),
        ([1, 3, np.nan], [3], 1, 1, 2, np.nanmax, LeavesMissingValues()),
        ([1, 3, np.nan], [3], 1, 1, 2, np.nanmin, LeavesMissingValues()),
        ([1, 3, np.nan], [3], 1, 2, 1, np.nanmean, LeavesMissingValues()),
        ([1, 3, np.nan], [3], 1, 2, 1, np.nanmedian, LeavesMissingValues()),
        ([1, 3, np.nan], [3], 1, 2, 1, np.nanmax, LeavesMissingValues()),
        ([1, 3, np.nan], [3], 1, 2, 1, np.nanmin, LeavesMissingValues()),
        (
            [1, 3, np.nan],
            [3, np.nan],
            2,
            2,
            1,
            np.nanmean,
            LeavesMissingValues(),
        ),
        (
            [1, 3, np.nan],
            [3, np.nan],
            2,
            2,
            1,
            np.nanmedian,
            LeavesMissingValues(),
        ),
        (
            [1, 3, np.nan],
            [3, np.nan],
            2,
            2,
            1,
            np.nanmax,
            LeavesMissingValues(),
        ),
        (
            [1, 3, np.nan],
            [3, np.nan],
            2,
            2,
            1,
            np.nanmin,
            LeavesMissingValues(),
        ),
        # check if `nanmean` works when some seasons have missing values
        ([1, 3, np.nan], [3, 3], 2, 1, 2, np.nanmean, LeavesMissingValues()),
        (
            [1, 3, np.nan],
            [3, 3, 3],
            3,
            1,
            2,
            np.nanmean,
            LeavesMissingValues(),
        ),
        # check if `mean` works when some seasons have missing values
        (
            [1, 3, np.nan],
            [np.nan] * 2,
            2,
            1,
            2,
            np.mean,
            LeavesMissingValues(),
        ),
        (
            [1, 3, np.nan],
            [np.nan] * 3,
            3,
            1,
            2,
            np.mean,
            LeavesMissingValues(),
        ),
        # check if `nanmedian` works when some seasons have missing values
        ([1, 3, np.nan], [3, 3], 2, 1, 2, np.nanmedian, LeavesMissingValues()),
        (
            [1, 3, np.nan],
            [3, 3, 3],
            3,
            1,
            2,
            np.nanmedian,
            LeavesMissingValues(),
        ),
        # check if `nanmax` works when some seasons have missing values
        ([1, 3, np.nan], [3, 3], 2, 1, 2, np.nanmax, LeavesMissingValues()),
        ([1, 3, np.nan], [3, 3, 3], 3, 1, 2, np.nanmax, LeavesMissingValues()),
        # check if `nanmin` works when some seasons have missing values
        ([1, 3, np.nan], [3, 3], 2, 1, 2, np.nanmin, LeavesMissingValues()),
        ([1, 3, np.nan], [3, 3, 3], 3, 1, 2, np.nanmin, LeavesMissingValues()),
        # check if `mean` works when some seasons have missing values
        (
            [1, 3, np.nan],
            [np.nan] * 2,
            2,
            1,
            2,
            np.median,
            LeavesMissingValues(),
        ),
        (
            [1, 3, np.nan],
            [np.nan] * 3,
            3,
            1,
            2,
            np.median,
            LeavesMissingValues(),
        ),
        # check if `median` works when some seasons have missing values
        (
            [1, 3, np.nan],
            [np.nan] * 2,
            2,
            1,
            2,
            np.median,
            LeavesMissingValues(),
        ),
        (
            [1, 3, np.nan],
            [np.nan] * 3,
            3,
            1,
            2,
            np.median,
            LeavesMissingValues(),
        ),
        # check if `max` works when some seasons have missing values
        ([1, 3, np.nan], [np.nan] * 2, 2, 1, 2, np.max, LeavesMissingValues()),
        ([1, 3, np.nan], [np.nan] * 3, 3, 1, 2, np.max, LeavesMissingValues()),
        # check if `min` works when some seasons have missing values
        ([1, 3, np.nan], [np.nan] * 2, 2, 1, 2, np.min, LeavesMissingValues()),
        ([1, 3, np.nan], [np.nan] * 3, 3, 1, 2, np.min, LeavesMissingValues()),
    ],
)
def test_predictor(
    data,
    expected_output,
    prediction_length,
    season_length,
    num_seasons,
    agg_fun,
    imputation_method,
):
    prediction = get_prediction(
        data,
        prediction_length=prediction_length,
        season_length=season_length,
        num_seasons=num_seasons,
        agg_fun=agg_fun,
        imputation_method=imputation_method,
    )
    assert prediction.samples.shape == (1, prediction_length)

    np.testing.assert_equal(prediction.mean, expected_output)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest

from gluonts.mx.model.deepstate import DeepStateEstimator


@pytest.fixture()
def hyperparameters(dsinfo):
    return dict(
        ctx="cpu",
        epochs=1,
        learning_rate=1e-2,
        hybridize=False,
        num_cells=2,
        num_layers=1,
        context_length=2,
        past_length=dsinfo.prediction_length,
        cardinality=[1],
        use_feat_static_cat=False,
        num_batches_per_epoch=1,
        use_symbol_block_predictor=False,
    )


def test_accuracy(accuracy_test, hyperparameters):
    hyperparameters.update(num_batches_per_epoch=100)

    accuracy_test(DeepStateEstimator, hyperparameters, accuracy=0.77)


def test_serialize(serialize_test, hyperparameters):
    serialize_test(DeepStateEstimator, hyperparameters)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import tempfile
import zipfile
from pathlib import Path
from urllib import request

import pandas as pd
from gluonts.dataset import DatasetWriter
from gluonts.dataset.common import MetaData, TrainDatasets
from gluonts.dataset.repository._util import metadata


def generate_uber_dataset(
    dataset_path: Path,
    uber_freq: str,
    prediction_length: int,
    dataset_writer: DatasetWriter,
):
    subsets = {"daily": "1D", "hourly": "1h"}
    assert (
        uber_freq.lower() in subsets
    ), f"invalid uber_freq='{uber_freq}'. Allowed values: {subsets.keys()}"
    freq_setting = subsets[uber_freq.lower()]

    # download the dataset and read the data
    with tempfile.TemporaryDirectory() as dir_path:
        temp_dir_path = Path(dir_path)
        temp_zip_path = temp_dir_path / "uber-dataset.zip"
        uber_url_path = (
            "http://raw.githubusercontent.com/fivethirtyeight/"
            "uber-tlc-foil-response/master/uber-trip-data/"
            "uber-raw-data-janjune-15.csv.zip"
        )
        request.urlretrieve(uber_url_path, temp_zip_path)
        with zipfile.ZipFile(temp_zip_path) as zf:
            zf.extractall(path=temp_dir_path)
        uber_file_path = temp_dir_path / "uber-raw-data-janjune-15.csv"
        uber_df = pd.read_csv(
            uber_file_path,
            header=0,
            usecols=["Pickup_date", "locationID"],
            index_col=0,
        )

    # We divide the raw data according to locationID. Each json line represents
    # a time series of a loacationID. The targets are numbers of pickup-events
    # during a day or an hour.
    time_series_of_locations = list(uber_df.groupby(by="locationID"))

    train_data = []
    test_data = []
    for locationID, df in time_series_of_locations:
        df.sort_index()
        df.index = pd.to_datetime(df.index)

        count_series = df.resample(rule=freq_setting).size()
        start_time = pd.Timestamp(df.index[0]).strftime("%Y-%m-%d %X")
        target = count_series.values.tolist()
        feat_static_cat = [locationID]

        test_format_dict = {
            "start": start_time,
            "target": target,
            "feat_static_cat": feat_static_cat,
            "item_id": locationID,
        }
        test_data.append(test_format_dict)

        train_format_dict = {
            "start": start_time,
            "target": target[:-prediction_length],
            "feat_static_cat": feat_static_cat,
            "item_id": locationID,
        }
        train_data.append(train_format_dict)

    meta = MetaData(
        **metadata(
            cardinality=len(time_series_of_locations),
            freq=freq_setting[1],
            prediction_length=prediction_length,
        )
    )

    dataset = TrainDatasets(metadata=meta, train=train_data, test=test_data)
    dataset.save(
        path_str=str(dataset_path), writer=dataset_writer, overwrite=True
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
import shutil
from pathlib import Path
from typing import Dict, Optional, Tuple

from gluonts.pydantic import BaseModel

from .dyn import install_and_restart
from .params import decode_sagemaker_parameters
from .nested_params import decode_nested_parameters


class DataConfig(BaseModel):
    ContentType: Optional[str] = None
    TrainingInputMode: Optional[str] = None
    S3DistributionType: Optional[str] = None
    RecordWrapperType: Optional[str] = None


class InpuDataConfig(BaseModel):
    __root__: Dict[str, DataConfig]

    def __getitem__(self, item):
        return self.__root__[item]

    def channels(self):
        return self.__root__

    def channel_names(self):
        return list(self.__root__.keys())


class TrainPaths:
    def __init__(self, base: Path = Path("/opt/ml")) -> None:
        self.base = base.expanduser().resolve()
        self.input = self.base / "input"
        self.config = self.base / "input" / "config"
        self.data = self.base / "input" / "data"
        self.model = self.base / "model"
        self.output = self.base / "output"
        self.failure = self.output / "failure"

        self.hyperparameters = self.config / "hyperparameters.json"
        self.inputdataconfig = self.config / "inputdataconfig.json"
        self.resourceconfig = self.config / "resourceconfig.json"

        self.config.mkdir(parents=True, exist_ok=True)
        self.data.mkdir(parents=True, exist_ok=True)
        self.model.mkdir(parents=True, exist_ok=True)
        self.output.mkdir(parents=True, exist_ok=True)


class TrainEnv:
    def __init__(self, path: Path = Path("/opt/ml")) -> None:
        self.path = TrainPaths(path)
        self.inputdataconfig = self._load_inputdataconfig()
        self.channels = self._load_channels()
        self.current_host = self._get_current_host()

        hyperparameters, env = self._load_hyperparameters()
        self.hyperparameters = hyperparameters
        self.env = env

    def install_dynamic(self):
        install_and_restart(self.channels.get("code"), self.path.base / "code")

    def copy_code_to_model(self):
        code = self.channels.get("code")
        if code is not None:
            dest = self.path.model / "code"

            if dest.is_dir():
                shutil.rmtree(dest)

            shutil.copytree(code, dest)

    def _load_inputdataconfig(self) -> Optional[InpuDataConfig]:
        if self.path.inputdataconfig.exists():
            return InpuDataConfig.parse_file(self.path.inputdataconfig)
        return None

    def _load_channels(self) -> Dict[str, Path]:
        """
        Lists the available channels in `/opt/ml/input/data`.

        Return:
        Dict of channel-names mapping to the corresponding path.

        When running in SageMaker, channels and are listed in
        `/opt/ml/config/inputdataconfig.json`. Thus, if this file is present,
        we take its content to determine which channels are available. To
        support a local development setup, we just list the contents of the
        data folder to get the available channels.
        """
        if self.inputdataconfig is not None:
            return {
                name: self.path.data / name
                for name in self.inputdataconfig.channel_names()
            }
        else:
            return {
                channel.name: channel for channel in self.path.data.iterdir()
            }

    def _get_current_host(self) -> str:
        if not self.path.resourceconfig.exists():
            return "local"
        else:
            with self.path.resourceconfig.open() as json_file:
                config = json.load(json_file)
                return config["current_host"]

    def _load_hyperparameters(self) -> Tuple[dict, Optional[dict]]:
        with self.path.hyperparameters.open() as json_file:
            raw = json.load(json_file)
            decoded = decode_sagemaker_parameters(raw)

            nested = decode_nested_parameters(decoded)
            hyperparameters = nested.get("", {})
            env = nested.get("env", None)
            return hyperparameters, env



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from dataclasses import dataclass, field
from typing import List, Union, Dict, Any, Optional
import logging
import math
import time

# Third-party imports
import mxnet.gluon.nn as nn
import mxnet as mx
from mxnet import gluon

# First-party imports
from gluonts.core.component import validated
from gluonts.mx.util import copy_parameters
from gluonts.pydantic import BaseModel, PrivateAttr

logger = logging.getLogger(__name__)


class Callback:
    """
    Abstract Callback base class.

    Callbacks control the training of the GluonTS trainer. To write a custom
    Callback, you can subclass Callback and overwrite one or more of the hook
    methods. Hook methods with boolean return value stop the training if False
    is returned.
    """

    def on_train_start(self, max_epochs: int) -> None:
        """
        Hook that is called prior to training. This is the very first hook to
        be called.

        Parameters
        ----------
        max_epochs
            The maximum number of epochs that training is running. The actual
            number of epochs may be fewer if another callback hook stops
            training early.
        """

    def on_network_initializing_end(
        self, training_network: nn.HybridBlock
    ) -> None:
        """
        Hook that is called prior to training, after the training network has
        been initialized. This is the first hook where the network is passed.

        Parameters
        ----------
        training_network
            The network that is being trained.
        """

    def on_train_epoch_start(self, training_network: nn.HybridBlock) -> None:
        """
        Hook that is called prior to each training epoch.

        Parameters
        ----------
        training_network
            The network that is being trained.
        """

    def on_validation_epoch_start(
        self, training_network: nn.HybridBlock
    ) -> None:
        """
        Hook that is called prior to each validation epoch. This hook is never
        called if no validation data is available during training.

        Parameters
        ----------
        training_network
            The network that is being trained.
        """

    def on_train_batch_end(self, training_network: nn.HybridBlock) -> bool:
        """
        Hook that is called after each training batch.

        Parameters
        ----------
        training_network
            The network that is being trained.

        Returns
        -------
        bool
            A boolean whether the training should continue. Defaults to `True`.
        """
        return True

    def on_validation_batch_end(
        self, training_network: nn.HybridBlock
    ) -> bool:
        """
        Hook that is called after each validation batch. This hook is never
        called if no validation data is available during training.

        Parameters
        ----------
        training_network
            The network that is being trained.

        Returns
        -------
        bool
            A boolean whether the training should continue. Defaults to `True`.
        """
        return True

    def on_train_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network: nn.HybridBlock,
        trainer: gluon.Trainer,
    ) -> bool:
        """
        Hook that is called after each training epoch. This method returns a
        boolean whether training should continue.

        Parameters
        ----------
        epoch_no
            The current epoch (the first epoch has `epoch_no = 0`).
        epoch_loss
            The loss that was recorded in the last epoch.
        training_network
            The network that is being trained.
        trainer
            The trainer which is running the training.

        Returns
        -------
        bool
            A boolean whether the training should continue. Defaults to `True`.
        """
        return True

    def on_validation_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network: nn.HybridBlock,
        trainer: gluon.Trainer,
    ) -> bool:
        """
        Hook that is called after each validation epoch. Similar to
        `on_train_epoch_end`, this method returns a boolean whether training
        should continue. Note that it is always called after
        `on_train_epoch_end` within a single epoch. If `on_train_epoch_end`
        returned `False`, this method will not be called.

        Parameters
        ----------
        epoch_no
            The current epoch (the first epoch has `epoch_no = 0`).
        epoch_loss
            The validation loss that was recorded in the last epoch.
        training_network
            The network that is being trained.
        trainer
            The trainer which is running the training.

        Returns
        -------
        bool
            A boolean whether the training should continue. Defaults to `True`.
        """
        return True

    def on_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network: nn.HybridBlock,
        trainer: gluon.Trainer,
        best_epoch_info: Dict[str, Any],
        ctx: mx.Context,
    ) -> bool:
        """
        Hook that is called after every epoch. As `on_train_epoch_end` and
        `on_validation_epoch_end`, it returns a boolean whether training should
        continue. This hook is always called after `on_train_epoch_end` and
        `on_validation_epoch_end`. It is called regardless of these hooks'
        return values.

        Parameters
        ----------
        epoch_no
            The current epoch (the first epoch has `epoch_no = 0`).
        epoch_loss
            The validation loss that was recorded in the last epoch if
            validation data was provided. The training loss otherwise.
        training_network
            The network that is being trained.
        trainer
            The trainer which is running the training.
        best_epoch_info
            Aggregate information about the best epoch. Contains keys
            `params_path`, `epoch_no` and `score`. The score is the best
            validation loss if validation data is provided or the best training
            loss otherwise.
        ctx
            The MXNet context used.

        Returns
        -------
        bool
            A boolean whether the training should continue. Defaults to `True`.
        """
        return True

    def on_train_end(
        self,
        training_network: nn.HybridBlock,
        temporary_dir: str,
        ctx: mx.context.Context = None,
    ) -> None:
        """
        Hook that is called after training is finished. This is the last hook
        to be called.

        Parameters
        ----------
        training_network
            The network that was trained.
        temporary_dir
            The directory where model parameters are logged throughout
            training.
        ctx
            An MXNet context used.
        """


class CallbackList(Callback):
    """
    Used to chain a list of callbacks to one Callback. Boolean hook methods are
    logically joined with AND, meaning that if at least one callback method
    returns False, the training is stopped.

    Attributes
    ----------
    callbacks
        A list of gluonts.mx.trainer.callback.Callback's.
    """

    @validated()
    def __init__(self, callbacks: List[Callback]):
        self.callbacks = callbacks

    def _exec(
        self, function_name: str, *args: Any, **kwargs: Any
    ) -> Union[List[bool], List[None]]:
        return [
            getattr(callback, function_name)(*args, **kwargs)
            for callback in self.callbacks
        ]

    def on_train_start(self, *args: Any, **kwargs: Any) -> None:
        self._exec("on_train_start", *args, **kwargs)

    def on_network_initializing_end(self, *args: Any, **kwargs: Any) -> None:
        self._exec("on_network_initializing_end", *args, **kwargs)

    def on_train_epoch_start(self, *args: Any, **kwargs: Any) -> None:
        self._exec("on_train_epoch_start", *args, **kwargs)

    def on_validation_epoch_start(self, *args: Any, **kwargs: Any) -> None:
        self._exec("on_validation_epoch_start", *args, **kwargs)

    def on_train_batch_end(self, *args: Any, **kwargs: Any) -> bool:
        return all(self._exec("on_train_batch_end", *args, **kwargs))

    def on_validation_batch_end(self, *args: Any, **kwargs: Any) -> bool:
        return all(self._exec("on_validation_batch_end", *args, **kwargs))

    def on_train_epoch_end(self, *args: Any, **kwargs: Any) -> bool:
        return all(self._exec("on_train_epoch_end", *args, **kwargs))

    def on_validation_epoch_end(self, *args: Any, **kwargs: Any) -> bool:
        return all(self._exec("on_validation_epoch_end", *args, **kwargs))

    def on_epoch_end(self, *args: Any, **kwargs: Any) -> bool:
        return all(self._exec("on_epoch_end", *args, **kwargs))

    def on_train_end(self, *args: Any, **kwargs: Any) -> None:
        self._exec("on_train_end", *args, **kwargs)


class TrainingHistory(Callback):
    @validated()
    def __init__(self):
        self.loss_history = []
        self.validation_loss_history = []

    def on_train_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network: nn.HybridBlock,
        trainer: gluon.Trainer,
    ) -> bool:
        self.loss_history.append(epoch_loss)
        return True

    def on_validation_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network: nn.HybridBlock,
        trainer: gluon.Trainer,
    ) -> bool:
        self.validation_loss_history.append(epoch_loss)
        return True


class TerminateOnNaN(Callback):
    def on_train_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network: nn.HybridBlock,
        trainer: gluon.Trainer,
    ) -> bool:
        if math.isnan(epoch_loss):
            logging.warning(
                "TerminateOnNaN Callback initiated stop of training at epoch"
                f" {epoch_no}."
            )
            return False
        return True


class WarmStart(Callback):
    @validated()
    def __init__(self, predictor):
        self.predictor = predictor

    def on_network_initializing_end(
        self, training_network: nn.HybridBlock
    ) -> None:
        copy_parameters(self.predictor.prediction_net, training_network)


@dataclass
class _Timer:
    duration: float
    _start: Optional[float] = field(init=False, default=None)

    def start(self):
        assert self._start is None

        self._start = time.time()

    def remaining(self) -> float:
        assert self._start is not None

        return max(self._start - time.time(), 0.0)

    def is_running(self) -> bool:
        return self.remaining() > 0


class TrainingTimeLimit(BaseModel, Callback):
    """
    Limit time spent for training.

    This is useful when ensuring that training for a given model doesn't
    exceed a budget, for example when doing AutoML.

    If `stop_within_epoch` is set to true, training can be stopped after
    each batch, otherwise it stops after the end of the epoch.
    """

    time_limit: float
    stop_within_epoch: bool = False
    _timer: _Timer = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._timer = _Timer(self.time_limit)

    def on_train_start(self, max_epochs: int) -> None:
        self._timer.start()

    def on_train_batch_end(self, training_network: nn.HybridBlock) -> bool:
        if self.stop_within_epoch:
            return self._timer.is_running()

        return True

    def on_epoch_end(
        self,
        epoch_no: int,
        epoch_loss: float,
        training_network: nn.HybridBlock,
        trainer: gluon.Trainer,
        best_epoch_info: Dict[str, Any],
        ctx: mx.Context,
    ) -> bool:
        return self._timer.is_running()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
import sys
from collections import Counter
from contextlib import contextmanager
from itertools import islice
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Any, List

import numpy as np
import pytest
from toolz import take

from gluonts.core.component import equals
from gluonts.dataset.artificial import constant_dataset
from gluonts.dataset.common import DataEntry, DataBatch, FileDataset
from gluonts.dataset.field_names import FieldName
from gluonts.itertools import Cyclic
from gluonts.dataset.loader import (
    Batch,
    TrainDataLoader,
    ValidationDataLoader,
    InferenceDataLoader,
    as_stacked_batches,
)
from gluonts.transform import (
    InstanceSampler,
    InstanceSplitter,
    MapTransformation,
)

from gluonts.testutil.batchify import batchify


class WriteIsTrain(MapTransformation):
    def map_transform(self, data: DataEntry, is_train: bool) -> DataEntry:
        data["is_train"] = is_train
        return data


class ExactlyOneSampler(InstanceSampler):
    def __call__(self, ts: np.ndarray) -> np.ndarray:
        a, b = self._get_bounds(ts)
        window_size = b - a + 1
        assert window_size > 0
        return np.array([a])


@contextmanager
def default_list_dataset():
    yield constant_dataset()[1]


@contextmanager
def default_file_dataset():
    with TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)

        with open(temp_dir / "file.json", "w") as fp:
            for item_id in range(10):
                json.dump(
                    {
                        "item_id": item_id,
                        "start": "2021-01-01 00:00:00",
                        "target": [1.0, 2.0, 3.0, 4.0, 5.0],
                    },
                    fp,
                )
                fp.write("\n")

        yield FileDataset(temp_dir, "H")


def default_transformation():
    return WriteIsTrain() + InstanceSplitter(
        target_field=FieldName.TARGET,
        is_pad_field=FieldName.IS_PAD,
        start_field=FieldName.START,
        forecast_start_field=FieldName.FORECAST_START,
        instance_sampler=ExactlyOneSampler(),
        past_length=10,
        future_length=5,
        dummy_value=1.0,
    )


def count_item_ids(batches: List[DataBatch]) -> Dict[Any, int]:
    return Counter(
        item_id for batch in batches for item_id in batch[FieldName.ITEM_ID]
    )


@pytest.mark.parametrize(
    "dataset_context",
    (
        [
            default_list_dataset,
            default_file_dataset,
        ]
        if not sys.platform.startswith("win")
        else [default_list_dataset]
    ),
)
def test_training_data_loader(dataset_context):
    with dataset_context() as dataset:
        dataset_length = len(dataset)

        batch_size = 4

        dl = TrainDataLoader(
            dataset=dataset,
            transform=default_transformation(),
            batch_size=batch_size,
            stack_fn=batchify,
        )

        num_epochs = 20
        epoch_length = 2

        passes_through_dataset = int(
            (num_epochs * epoch_length * batch_size) / dataset_length
        )

        # these are to make sure that the test makes sense:
        # we want to go over the dataset multiple times
        assert passes_through_dataset >= 10
        # we want each epoch to be shorter than the dataset
        assert epoch_length * batch_size < dataset_length

        batches = []

        for epoch in range(num_epochs):
            for batch in islice(dl, epoch_length):
                assert all(batch["is_train"])
                batches.append(batch)

        counter = count_item_ids(batches)

        for entry in dataset:
            assert counter[entry[FieldName.ITEM_ID]] >= 1


@pytest.mark.parametrize(
    "dataset_context",
    [
        default_list_dataset,
        default_file_dataset,
    ],
)
def test_validation_data_loader(dataset_context):
    with dataset_context() as dataset:
        dl = ValidationDataLoader(
            dataset=dataset,
            transform=default_transformation(),
            batch_size=4,
            stack_fn=batchify,
        )

        for _ in range(3):
            batches = list(dl)

            for batch in batches:
                assert all(batch["is_train"])

            counter = count_item_ids(batches)

            for entry in dataset:
                assert counter[entry[FieldName.ITEM_ID]] == 1


@pytest.mark.parametrize(
    "dataset_context",
    [
        default_list_dataset,
        default_file_dataset,
    ],
)
def test_inference_data_loader(dataset_context):
    with dataset_context() as dataset:
        dl = InferenceDataLoader(
            dataset=dataset,
            transform=default_transformation(),
            batch_size=4,
            stack_fn=batchify,
        )

        batches = list(dl)

        for batch in batches:
            assert not any(batch["is_train"])

        counter = count_item_ids(batches)

        for entry in dataset:
            assert counter[entry[FieldName.ITEM_ID]] == 1


def test_equals_batch():
    assert equals(Batch(batch_size=10), Batch(batch_size=10))
    assert not equals(Batch(batch_size=10), Batch(batch_size=100))


def test_as_stacked_batches():
    step = 10
    data = [
        {"x": np.arange(start, start + step)} for start in range(0, 100, step)
    ]

    stream = as_stacked_batches(data, batch_size=2)

    for _ in range(3):
        batches = list(take(2, stream))
        assert np.array_equal(batches[0]["x"], np.arange(0, 20).reshape(2, 10))
        assert np.array_equal(
            batches[1]["x"], np.arange(20, 40).reshape(2, 10)
        )


def test_as_stacked_batches_iter():
    step = 10
    data = iter(
        [
            {"x": np.arange(start, start + step)}
            for start in range(0, 100, step)
        ]
    )

    stream = as_stacked_batches(data, batch_size=2)

    batches = list(take(2, stream))
    assert np.array_equal(batches[0]["x"], np.arange(0, 20).reshape(2, 10))
    assert np.array_equal(batches[1]["x"], np.arange(20, 40).reshape(2, 10))

    batches = list(take(2, stream))
    assert np.array_equal(batches[0]["x"], np.arange(40, 60).reshape(2, 10))
    assert np.array_equal(batches[1]["x"], np.arange(60, 80).reshape(2, 10))

    batches = list(take(2, stream))
    assert len(batches) == 1
    assert np.array_equal(batches[0]["x"], np.arange(80, 100).reshape(2, 10))


def test_as_stacked_batches_iter_num_batches():
    step = 10
    data = iter(
        [
            {"x": np.arange(start, start + step)}
            for start in range(0, 100, step)
        ]
    )

    stream = as_stacked_batches(data, batch_size=2, num_batches_per_epoch=3)

    batches = list(stream)
    assert len(batches) == 3
    assert np.array_equal(batches[0]["x"], np.arange(0, 20).reshape(2, 10))
    assert np.array_equal(batches[1]["x"], np.arange(20, 40).reshape(2, 10))
    assert np.array_equal(batches[2]["x"], np.arange(40, 60).reshape(2, 10))

    batches = list(stream)
    assert len(batches) == 2
    assert np.array_equal(batches[0]["x"], np.arange(60, 80).reshape(2, 10))
    assert np.array_equal(batches[1]["x"], np.arange(80, 100).reshape(2, 10))

    assert len(list(stream)) == 0


def test_as_stacked_batches_num_batches_iter_cycle():
    step = 10
    data = iter(
        Cyclic(
            [
                {"x": np.arange(start, start + step)}
                for start in range(0, 100, step)
            ]
        )
    )

    stream = as_stacked_batches(data, batch_size=2, num_batches_per_epoch=3)

    batches = list(stream)
    assert len(batches) == 3
    assert np.array_equal(batches[0]["x"], np.arange(0, 20).reshape(2, 10))
    assert np.array_equal(batches[1]["x"], np.arange(20, 40).reshape(2, 10))
    assert np.array_equal(batches[2]["x"], np.arange(40, 60).reshape(2, 10))

    batches = list(stream)
    assert len(batches) == 3
    assert np.array_equal(batches[0]["x"], np.arange(60, 80).reshape(2, 10))
    assert np.array_equal(batches[1]["x"], np.arange(80, 100).reshape(2, 10))
    assert np.array_equal(batches[2]["x"], np.arange(0, 20).reshape(2, 10))

    batches = list(stream)
    assert len(batches) == 3
    assert np.array_equal(batches[0]["x"], np.arange(20, 40).reshape(2, 10))
    assert np.array_equal(batches[1]["x"], np.arange(40, 60).reshape(2, 10))
    assert np.array_equal(batches[2]["x"], np.arange(60, 80).reshape(2, 10))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
import logging
from typing import Any, Optional, Type, Union

import gluonts
from gluonts.core import fqname_for
from gluonts.dataset.common import Dataset
from gluonts.evaluation import Evaluator, backtest
from gluonts.model.estimator import Estimator, IncrementallyTrainable
from gluonts.model.forecast import Quantile
from gluonts.model.predictor import Predictor
from gluonts.itertools import maybe_len
from gluonts.transform import FilterTransformation

from .env import TrainEnv
from .util import invoke_with

logger = logging.getLogger(__name__)


def log_metric(metric: str, value: Any) -> None:
    logger.info(f"gluonts[{metric}]: {value}")


def log_version(forecaster_type):
    name = fqname_for(forecaster_type)
    version = forecaster_type.__version__

    logger.info(f"Using gluonts v{gluonts.__version__}")
    logger.info(f"Using forecaster {name} v{version}")


def run_train_and_test(
    env: TrainEnv, forecaster_type: Type[Union[Estimator, Predictor]]
) -> None:
    log_version(forecaster_type)

    logger.info(
        "Using the following data channels: %s", ", ".join(env.datasets)
    )

    forecaster = forecaster_type.from_inputs(
        env.datasets["train"], **env.hyperparameters
    )
    logger.info(
        "The forecaster can be reconstructed with the following expression: "
        f"{forecaster}"
    )

    if isinstance(forecaster, Predictor):
        predictor = forecaster
    else:
        predictor = run_train(
            forecaster=forecaster,
            train_dataset=env.datasets["train"],
            validation_dataset=env.datasets.get("validation"),
            hyperparameters=env.hyperparameters,
            from_predictor=env.datasets.get("model"),
        )

    predictor.serialize(env.path.model)

    if "test" in env.datasets:
        run_test(env, predictor, env.datasets["test"], env.hyperparameters)


def run_train(
    forecaster: Estimator,
    train_dataset: Dataset,
    hyperparameters: dict,
    validation_dataset: Optional[Dataset],
    from_predictor: Optional[Predictor],
) -> Predictor:
    num_workers = (
        int(hyperparameters["num_workers"])
        if "num_workers" in hyperparameters
        else None
    )
    shuffle_buffer_length = (
        int(hyperparameters["shuffle_buffer_length"])
        if "shuffle_buffer_length" in hyperparameters
        else None
    )
    num_prefetch = (
        int(hyperparameters["num_prefetch"])
        if "num_prefetch" in hyperparameters
        else None
    )

    if from_predictor is not None:
        assert isinstance(forecaster, IncrementallyTrainable), (
            "The model provided does not implement the "
            "IncrementallyTrainable protocol"
        )
        return invoke_with(
            forecaster.train_from,
            from_predictor,
            training_data=train_dataset,
            validation_data=validation_dataset,
            num_workers=num_workers,
            num_prefetch=num_prefetch,
            shuffle_buffer_length=shuffle_buffer_length,
        )

    return invoke_with(
        forecaster.train,
        training_data=train_dataset,
        validation_data=validation_dataset,
        num_workers=num_workers,
        num_prefetch=num_prefetch,
        shuffle_buffer_length=shuffle_buffer_length,
    )


def run_test(
    env: TrainEnv,
    predictor: Predictor,
    test_dataset: Dataset,
    hyperparameters: dict,
) -> None:
    len_original = maybe_len(test_dataset)

    test_dataset = FilterTransformation(
        lambda x: x["target"].shape[-1] > predictor.prediction_length
    ).apply(test_dataset)

    len_filtered = len(test_dataset)

    if len_original is not None and len_original > len_filtered:
        logger.warning(
            "Not all time series in the test-channel have "
            "enough data to be used for evaluation. Proceeding with "
            f"{len_filtered}/{len_original} "
            f"(~{int(len_filtered / len_original * 100)}%) items."
        )

    forecast_it, ts_it = backtest.make_evaluation_predictions(
        dataset=test_dataset, predictor=predictor, num_samples=100
    )

    test_quantiles = (
        [
            Quantile.parse(quantile).name
            for quantile in hyperparameters["test_quantiles"]
        ]
        if "test_quantiles" in hyperparameters
        else None
    )

    if test_quantiles is not None:
        logger.info(f"Using quantiles `{test_quantiles}` for evaluation.")
        evaluator = Evaluator(quantiles=test_quantiles)
    else:
        evaluator = Evaluator()

    agg_metrics, item_metrics = evaluator(
        ts_iterator=ts_it,
        fcst_iterator=forecast_it,
        num_series=len(test_dataset),
    )

    # we only log aggregate metrics for now as item metrics may be very large
    for name, score in agg_metrics.items():
        logger.info(f"#test_score ({env.current_host}, {name}): {score}")

    # store metrics
    with open(env.path.model / "agg_metrics.json", "w") as agg_metric_file:
        json.dump(agg_metrics, agg_metric_file)
    with open(env.path.model / "item_metrics.csv", "w") as item_metrics_file:
        item_metrics.to_csv(item_metrics_file, index=False)
